import dataclasses
import logging
import os
import re
from collections.abc import Iterable
from dataclasses import dataclass
from functools import cached_property
from typing import ClassVar, Optional
from urllib.parse import urlparse, ParseResult

from databricks.labs.blueprint.installation import Installation
from databricks.labs.lsql.backends import SqlBackend
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import NotFound
from databricks.sdk.service.catalog import ExternalLocationInfo
from databricks.sdk.dbutils import FileInfo
from databricks.labs.ucx.framework.crawlers import CrawlerBase
from databricks.labs.ucx.framework.utils import escape_sql_identifier
from databricks.labs.ucx.hive_metastore.tables import Table, TablesCrawler

logger = logging.getLogger(__name__)

_EXTERNAL_FILE_LOCATION_SCHEMES = ("s3", "s3a", "s3n", "gcs", "abfss")


@dataclass
class ExternalLocation:
    location: str
    table_count: int


# Union with string is invalid syntax: "LocationTrie" | None
OptionalLocationTrie = Optional["LocationTrie"]  # pylint: disable=consider-alternative-union-syntax


@dataclass
class LocationTrie:
    """A trie datastructure to search locations.

    Used to find overlapping locations.
    """

    key: str = ""
    parent: OptionalLocationTrie = dataclasses.field(repr=False, default=None)
    children: dict[str, "LocationTrie"] = dataclasses.field(default_factory=dict)
    tables: list[Table] = dataclasses.field(repr=False, default_factory=list)

    @cached_property
    def _path(self) -> list[str]:
        """The path to traverse to get to the current node."""
        parts = []
        current: LocationTrie | None = self
        while current:
            parts.append(current.key)
            current = current.parent
        return list(reversed(parts))[1:]

    @property
    def location(self) -> str | None:
        if not self.is_valid():
            return None
        try:
            scheme, netloc, *path = self._path
            return f"{scheme}://{netloc}/{'/'.join(path)}".rstrip("/")
        except ValueError:
            return None

    @classmethod
    def _parse_location(cls, location: str | None) -> list[str]:
        if not location:
            return []
        location = ExternalLocations.clean_location(location)
        parse_result = cls._parse_url(location)
        if not parse_result:
            return []
        parts = [parse_result.scheme, parse_result.netloc]
        for part in parse_result.path.split("/"):
            if not part:
                continue  # remove empty strings
            parts.append(part)
        return parts

    @staticmethod
    def _parse_url(location: str) -> ParseResult | None:
        parse_result = urlparse(location)
        if parse_result.scheme == 'jdbc':
            jdbc_path = parse_result.path.split('://')
            if len(jdbc_path) != 2:
                return None
            netloc, path = jdbc_path[1].split('/', 1)
            parse_result = ParseResult(
                scheme=f'{parse_result.scheme}:{jdbc_path[0]}',
                netloc=netloc,
                path=path,
                params='',
                query='',
                fragment='',
            )
        return parse_result

    def insert(self, table: Table) -> None:
        current = self
        for part in self._parse_location(table.location):
            if part not in current.children:
                parent = current
                current = LocationTrie(part, parent)
                parent.children[part] = current
                continue
            current = current.children[part]
        current.tables.append(table)

    def find(self, table: Table) -> OptionalLocationTrie:
        current = self
        for part in self._parse_location(table.location):
            if part not in current.children:
                return None
            current = current.children[part]
        return current

    def is_valid(self) -> bool:
        """A valid location has a scheme and netloc; the path is optional."""
        if len(self._path) < 2:
            return False
        scheme, netloc, *_ = self._path
        if scheme.startswith('jdbc:') and len(netloc) > 0:
            return True
        return scheme in _EXTERNAL_FILE_LOCATION_SCHEMES and len(netloc) > 0

    def is_jdbc(self) -> bool:
        if not self.is_valid():
            return False
        return self._path[0].startswith('jdbc:')

    def all_tables(self) -> Iterable[Table]:
        for node in self:
            yield from node.tables

    def has_children(self):
        return len(self.children) > 0

    def __iter__(self):
        if self.is_valid():
            yield self
        for child in self.children.values():
            yield from child


class ExternalLocations(CrawlerBase[ExternalLocation]):
    _prefix_size: ClassVar[list[int]] = [1, 12]

    def __init__(
        self,
        ws: WorkspaceClient,
        sql_backend: SqlBackend,
        schema: str,
        tables_crawler: TablesCrawler,
        mounts_crawler: 'MountsCrawler',
        *,
        enable_hms_federation: bool = False,
    ):
        super().__init__(sql_backend, "hive_metastore", schema, "external_locations", ExternalLocation)
        self._ws = ws
        self._tables_crawler = tables_crawler
        self._mounts_crawler = mounts_crawler
        self._enable_hms_federation = enable_hms_federation

    @cached_property
    def _mounts_snapshot(self) -> list['Mount']:
        """Returns all mounts, sorted by longest prefixes first."""
        return sorted(self._mounts_crawler.snapshot(), key=lambda _: (len(_.name), _.name), reverse=True)

    @staticmethod
    def clean_location(location: str) -> str:
        # remove the s3a scheme and replace it with s3 as these can be considered the same and will be treated as such
        # Having s3a and s3 as separate locations will cause issues when trying to find overlapping locations
        return re.sub(r"^s3a:/", r"s3:/", location).rstrip("/")

    def external_locations_with_root(self) -> Iterable[ExternalLocation]:
        """
        Produces a list of external locations with the DBFS root location appended to the list.
        Utilizes the snapshot method.
        Used for HMS Federation.

        Yields:
                Iterable[Result]: Combination of all the external locations and the DBFS root location
        """

        yield from self.snapshot()
        dbfs_root = self._get_dbfs_root()
        if dbfs_root:
            yield dbfs_root

    def _get_dbfs_root(self) -> ExternalLocation | None:
        """
        Get the root location of the DBFS only if HMS Fed is enabled.
        Utilizes an undocumented Databricks API call

        Returns:
            Cloud storage root location for dbfs

        """
        if not self._enable_hms_federation:
            return None
        logger.debug("Retrieving DBFS root location")
        try:
            response = self._ws.api_client.do("GET", "/api/2.0/dbfs/resolve-path", query={"path": "dbfs:/"})
            if isinstance(response, dict):
                resolved_path = response.get("resolved_path")
                if resolved_path:
                    path = f"{self.clean_location(resolved_path)}/user/hive/warehouse"
                    return ExternalLocation(path, 0)
        except NotFound:
            # Couldn't retrieve the DBFS root location
            logger.warning("DBFS root location not found")
            return None
        return None

    def _external_locations(self) -> Iterable[ExternalLocation]:
        trie = LocationTrie()
        for table in self._tables_crawler.snapshot():
            table = self._resolve_location(table)
            if not table.location:
                continue
            trie.insert(table)
        queue = list(trie.children.values())
        external_locations = []
        while queue:
            curr = queue.pop()
            num_children = len(curr.children)  # 0 - take parent
            if curr.location and (num_children > 1 or num_children == 0):
                # Checking if the parent location is a valid location for external location.
                # If the table location is a leaf location (a foldermore than 2 levels from the root)
                # the parent folder will be considered as external location.
                # If the table location is a root location (a folder at the root level) it will be considered as
                # external location.

                if (
                    curr.parent and curr.parent.is_valid() and num_children == 0 and not curr.is_jdbc()
                ):  # one table having the prefix
                    curr = curr.parent
                if not curr.location:
                    continue
                external_location = ExternalLocation(curr.location, len(list(curr.all_tables())))
                external_locations.append(external_location)
                continue
            queue.extend(curr.children.values())
        return sorted(external_locations, key=lambda _: _.location)

    def _resolve_location(self, table: Table) -> Table:
        location = table.location
        if not location:
            return table
        location = self._resolve_jdbc(table)
        location = self.resolve_mount(location)
        return dataclasses.replace(table, location=location)

    def resolve_mount(self, location: str | None) -> str | None:
        if not location:
            return None
        if location.startswith('/dbfs'):
            location = 'dbfs:' + location[5:]  # convert FUSE path to DBFS path
        if not location.startswith('dbfs:'):
            return location  # not a mount, save some cycles
        for mount in self._mounts_snapshot:
            prefix = mount.as_scheme_prefix()
            if not location.startswith(prefix):
                continue
            logger.debug(f"Replacing location {prefix} with {mount.source} in {location}")
            location = location.replace(prefix, mount.source)
            return location
        logger.debug(f"Mount not found for location {location}. Skipping replacement.")
        return location

    def _resolve_jdbc(self, table: Table) -> str | None:
        location = table.location
        if not location or not table.storage_properties or not location.startswith('jdbc:'):
            return location
        pattern = r"(\w+)=(.*?)(?=\s*,|\s*\])"
        # Find all matches in the input string
        # Storage properties is of the format
        # "personalAccessToken=*********(redacted),
        #  httpPath=/sql/1.0/warehouses/65b52fb5bd86a7be, host=dbc-test1-aa11.cloud.databricks.com,
        #  dbtable=samples.nyctaxi.trips"
        matches = re.findall(pattern, table.storage_properties)
        # Create a dictionary from the matches
        result_dict = dict(matches)
        # Fetch the value of host from the newly created dict
        host = result_dict.get("host", "")
        port = result_dict.get("port", "")
        database = result_dict.get("database", "")
        httppath = result_dict.get("httpPath", "")
        provider = result_dict.get("provider", "")
        # currently supporting databricks and mysql external tables
        # add other jdbc types
        if "databricks" in location.lower():
            return f"jdbc:databricks://{host};httpPath={httppath}"
        if "mysql" in location.lower():
            return f"jdbc:mysql://{host}:{port}/{database}"
        if not provider == "":
            return f"jdbc:{provider.lower()}://{host}:{port}/{database}"
        return f"{location.lower()}/{host}:{port}/{database}"

    def _crawl(self) -> Iterable[ExternalLocation]:
        return self._external_locations()

    def _try_fetch(self) -> Iterable[ExternalLocation]:
        for row in self._fetch(f"SELECT * FROM {escape_sql_identifier(self.full_name)}"):
            yield ExternalLocation(*row)

    @staticmethod
    def _get_ext_location_definitions(missing_locations: list[ExternalLocation]) -> list:
        tf_script = []
        cnt = 1
        res_name = ""
        for loc in missing_locations:
            for scheme in _EXTERNAL_FILE_LOCATION_SCHEMES:
                prefix = f"{scheme}://"
                prefix_len = len(prefix)
                if not loc.location.startswith(prefix):
                    continue
                if prefix == "abfss://":
                    container_sep_loc = loc.location.index("@")
                    container_name = loc.location[prefix_len:container_sep_loc]
                    res_name = (
                        loc.location[container_sep_loc + 1 :]
                        .replace(".dfs.core.windows.net", "")
                        .rstrip("/")
                        .replace("/", "_")
                    )
                    res_name = f"{container_name}_{res_name}"
                else:
                    res_name = loc.location[prefix_len:].rstrip("/").replace("/", "_")
            if res_name == "":
                # if the cloud storage url doesn't match the above condition or incorrect (example wasb://)
                # dont generate tf script and ignore
                logger.warning(f"unsupported storage format {loc.location}")
                continue
            script = f'resource "databricks_external_location" "{res_name}" {{ \n'
            script += f'    name = "{res_name}"\n'
            script += f'    url  = "{loc.location.rstrip("/")}"\n'
            script += "    credential_name = databricks_storage_credential.<storage_credential_reference>.id\n"
            script += "}\n"
            tf_script.append(script)
            cnt += 1
        return tf_script

    def match_table_external_locations(self) -> tuple[dict[str, int], list[ExternalLocation]]:
        existing_locations = list(self._ws.external_locations.list())
        table_locations = self.snapshot()
        matching_locations: dict[str, int] = {}
        missing_locations = []
        for table_loc in table_locations:
            # external_location.list returns url without trailing "/" but ExternalLocation.snapshot
            # does so removing the trailing slash before comparing
            if not self._match_existing(table_loc, matching_locations, existing_locations):
                missing_locations.append(table_loc)
        return matching_locations, missing_locations

    @staticmethod
    def _match_existing(table_loc, matching_locations: dict[str, int], existing_locations: list[ExternalLocationInfo]):
        for uc_loc in existing_locations:
            if not uc_loc.url:
                continue
            if not uc_loc.name:
                continue
            uc_loc_path = uc_loc.url.lower()
            if uc_loc_path in table_loc.location.rstrip("/").lower():
                if uc_loc.name not in matching_locations:
                    matching_locations[uc_loc.name] = table_loc.table_count
                else:
                    matching_locations[uc_loc.name] = matching_locations[uc_loc.name] + table_loc.table_count
                return True
        return False

    def save_as_terraform_definitions_on_workspace(self, installation: Installation):
        matching_locations, missing_locations = self.match_table_external_locations()
        if len(matching_locations) > 0:
            logger.info("following external locations are already configured.")
            logger.info("sharing details of # tables that can be migrated for each location")
            for location, table_count in matching_locations.items():
                logger.info(f"{table_count} tables can be migrated using UC external location: {location}.")
        if len(missing_locations) > 0:
            logger.info("following external location need to be created.")
            for _ in missing_locations:
                logger.info(f"{_.table_count} tables can be migrated using external location {_.location}.")
            buffer = []
            for script in self._get_ext_location_definitions(missing_locations):
                buffer.append(script)
            return installation.upload('external_locations.tf', ("\n".join(buffer)).encode('utf8'))
        logger.info("no additional external location to be created.")
        return None


@dataclass(unsafe_hash=True)
class Mount:
    name: str
    source: str

    def as_scheme_prefix(self) -> str:
        return f'dbfs:{self.name}'  # dbfs:/mnt/mount-name


class MountsCrawler(CrawlerBase[Mount]):
    def __init__(
        self,
        sql_backend: SqlBackend,
        ws: WorkspaceClient,
        inventory_database: str,
    ):
        super().__init__(sql_backend, "hive_metastore", inventory_database, "mounts", Mount)
        self._dbutils = ws.dbutils

    @staticmethod
    def _deduplicate_mounts(mounts: list) -> list:
        seen = set()
        deduplicated_mounts = []
        for obj in mounts:
            if "dbfsreserved" in obj.source.lower():
                obj_tuple = Mount("/Volume", obj.source)
            else:
                obj_tuple = Mount(obj.name, obj.source)
            if obj_tuple not in seen:
                seen.add(obj_tuple)
                deduplicated_mounts.append(obj)
        return deduplicated_mounts

    @cached_property
    def _jvm(self):
        # pylint: disable=import-error,import-outside-toplevel,broad-exception-caught
        try:
            from pyspark.sql.session import SparkSession  # type: ignore[import-not-found]

            spark = SparkSession.builder.getOrCreate()
            return spark._jvm  # pylint: disable=protected-access
        except Exception as err:
            logger.warning(f"Cannot create Py4j proxy: {err}")
            return None

    def _resolve_dbfs_root(self) -> Mount | None:
        # TODO: Consider deprecating this method and rely on the new API call
        # pylint: disable=broad-exception-caught,too-many-try-statements
        try:
            jvm = self._jvm
            if not jvm:
                return None
            uri = jvm.java.net.URI
            some = jvm.scala.Some
            hms_fed_dbfs_utils = jvm.com.databricks.sql.managedcatalog.connections.HmsFedDbfsUtils
            root_location_opt = hms_fed_dbfs_utils.resolveDbfsPath(some(uri("dbfs:/user/hive/warehouse")))
            if root_location_opt.isDefined():
                source: str = root_location_opt.get().toString()
                source = source.removesuffix('user/hive/warehouse')
                return Mount("/", source)
            return None
        except Exception as err:
            logger.warning(f"Failed to resolve DBFS root location: {err}")
            return None

    def _crawl(self) -> Iterable[Mount]:
        mounts = []
        try:
            for mount_point, source, _ in self._dbutils.fs.mounts():
                mounts.append(Mount(mount_point, source))
        except Exception as error:  # pylint: disable=broad-except
            if "com.databricks.backend.daemon.dbutils.DBUtilsCore.mounts() is not whitelisted" in str(error):
                logger.warning(
                    "dbutils.fs.mounts() is not whitelisted. Skipping mount point discovery."
                    "Please make sure you run the assessment workflow."
                )
        return self._deduplicate_mounts(mounts)

    def _try_fetch(self) -> Iterable[Mount]:
        for row in self._fetch(f"SELECT * FROM {escape_sql_identifier(self.full_name)}"):
            yield Mount(*row)


@dataclass
class TableInMount:
    format: str
    is_partitioned: bool

    def __post_init__(self) -> None:
        if isinstance(self.format, str):  # Should not happen according to type hint, still safer
            self.format = self.format.upper()


class TablesInMounts(CrawlerBase[Table]):
    """Experimental scanner for tables that can be found on mounts.

    This crawler was developed with a specific use-case in mind and isn't currently in use. It does not conform to the
    design of other crawlers. In particular:
     - It depends on the `tables` inventory, but without verifying the tables crawler has run.
     - Rather than have its own table it will blindly overwrite the existing content of the tables inventory.
    """

    TABLE_IN_MOUNT_DB = "mounted_"

    def __init__(
        self,
        sql_backend: SqlBackend,
        ws: WorkspaceClient,
        inventory_database: str,
        mounts_crawler: MountsCrawler,
        include_mounts: list[str] | None = None,
        exclude_paths_in_mount: list[str] | None = None,
        include_paths_in_mount: list[str] | None = None,
    ):
        super().__init__(sql_backend, "hive_metastore", inventory_database, "tables", Table)
        self._dbutils = ws.dbutils
        self._mounts_crawler = mounts_crawler
        self._include_mounts = include_mounts
        self._ws = ws
        self._include_paths_in_mount = include_paths_in_mount

        irrelevant_patterns = {'_SUCCESS', '_committed_', '_started_'}
        if exclude_paths_in_mount:
            irrelevant_patterns.update(exclude_paths_in_mount)
        self._fiter_paths = irrelevant_patterns

    def snapshot(self, *, force_refresh: bool = False) -> list[Table]:
        if not force_refresh:
            msg = "This crawler only supports forced refresh; refer to source implementation for details."
            raise NotImplementedError(msg)
        return list(super().snapshot(force_refresh=force_refresh))

    def _crawl(self) -> list[Table]:
        logger.debug(f"[{self.full_name}] fetching {self._table} inventory")
        try:
            cached_results = list(self._try_fetch())
        except NotFound:
            # This happens when the table crawler hasn't run yet, and is arguably incorrect:
            # rather than pretending there are no tables it should instead trigger a crawl.
            cached_results = []
        table_paths = self._get_tables_paths_from_assessment(cached_results)
        logger.debug(f"[{self.full_name}] crawling new batch for {self._table}")
        loaded_records = list(self._crawl_tables(table_paths))
        if cached_results:
            loaded_records = [*loaded_records, *cached_results]
        return loaded_records

    def _try_fetch(self) -> Iterable[Table]:
        """Tries to load table information from the database or throws TABLE_OR_VIEW_NOT_FOUND error"""
        for row in self._fetch(
            f"SELECT * FROM {escape_sql_identifier(self.full_name)} WHERE NOT STARTSWITH(database, '{self.TABLE_IN_MOUNT_DB}')"
        ):
            yield Table(*row)

    @staticmethod
    def _get_tables_paths_from_assessment(loaded_records: Iterable[Table]) -> dict[str, str]:
        seen = {}
        for rec in loaded_records:
            if not rec.location:
                continue
            seen[rec.location] = rec.key
        return seen

    def _crawl_tables(self, table_paths_from_assessment: dict[str, str]) -> list[Table]:
        all_mounts = self._mounts_crawler.snapshot()
        all_tables = []
        for mount in all_mounts:
            if self._include_mounts and mount.name not in self._include_mounts:
                logger.info(f"Filtering mount {mount.name}")
                continue
            if self._include_paths_in_mount:
                table_paths = {}
                for path in self._include_paths_in_mount:
                    table_paths.update(self._find_delta_log_folders(path))
            else:
                table_paths = self._find_delta_log_folders(mount.name)

            for path, entry in table_paths.items():
                guess_table = os.path.basename(path)
                table_location = self._get_table_location(mount, path)

                # A table in mount may have already been identified by the assessment job because they're on the current workspace HMS
                # We filter those tables as we give better support to migrate those tables to UC
                if table_location in table_paths_from_assessment:
                    logger.info(
                        f"Path {path} is identified as a table in mount, but is present in current workspace as a registered table {table_paths_from_assessment[table_location]}"
                    )
                    continue
                if path in table_paths_from_assessment:
                    logger.info(
                        f"Path {path} is identified as a table in mount, but is present in current workspace as a registered table {table_paths_from_assessment[path]}"
                    )
                    continue
                table = Table(
                    catalog="hive_metastore",
                    database=f"{self.TABLE_IN_MOUNT_DB}{mount.name.replace('/mnt/', '').replace('/', '_')}",
                    name=guess_table,
                    object_type="EXTERNAL",
                    table_format=entry.format,
                    location=table_location,
                    is_partitioned=entry.is_partitioned,
                )
                all_tables.append(table)
        logger.info(f"Found a total of {len(all_tables)} tables in mount points")
        return all_tables

    @staticmethod
    def _get_table_location(mount: Mount, path: str) -> str:
        """
        There can be different cases for mounts:
            - Mount(name='/mnt/things/a', source='abfss://things@labsazurethings.dfs.core.windows.net/a')
            - Mount(name='/mnt/mount' source='abfss://container@dsss.net/')
            Both must return the complete source with a forward slash in the end
        """
        if mount.source.endswith("/"):
            return path.replace(f"dbfs:{mount.name}/", mount.source)
        return path.replace(f"dbfs:{mount.name}", mount.source)

    def _find_delta_log_folders(
        self,
        root_dir: str,
        delta_log_folders: dict[str, TableInMount] | None = None,
    ) -> dict[str, TableInMount]:
        if delta_log_folders is None:
            delta_log_folders = {}
        logger.info(f"Listing {root_dir}")
        file_infos = self._dbutils.fs.ls(root_dir) or []
        for file_info in file_infos:
            if self._is_irrelevant(file_info.name) or file_info.path == root_dir:
                logger.debug(f"Path {file_info.path} is irrelevant")
                continue

            root_path = os.path.dirname(root_dir)
            parent_entry = delta_log_folders.get(root_path)
            table_in_mount = self._assess_path(file_info)

            if parent_entry:
                # if the root_path was already found to be partitioned, we can break to reduce redundant ops
                if delta_log_folders[root_path] == TableInMount(format="DELTA", is_partitioned=True):
                    break

                if (
                    # Happens when first folder was _delta_log and next folders are partitioned folder
                    parent_entry.format == "DELTA"
                    and self._is_partitioned(file_info.name)
                ) or (
                    # Happens when previous entries where partitioned folders and the current one is delta_log
                    parent_entry.is_partitioned
                    and table_in_mount
                    and table_in_mount.format == "DELTA"
                ):
                    delta_log_folders[root_path] = TableInMount(format="DELTA", is_partitioned=True)
                    logger.debug(f"Added DELTA table for {root_path} (partitioned delta)")

                self._recurse_dir_if_needed(file_info, delta_log_folders)

            elif self._is_partitioned(file_info.name):
                self._handle_partitioned_file(file_info.path, root_path, delta_log_folders)

            elif not table_in_mount:
                self._recurse_dir_if_needed(file_info, delta_log_folders)

            elif table_in_mount.format == "DELTA" and file_info.name == "_delta_log/":
                delta_log_folders[root_path] = table_in_mount
                logger.debug(f"Added {table_in_mount.format} table for {root_path} (normal delta)")

            else:
                delta_log_folders[root_path] = table_in_mount
                logger.debug(f"Added {table_in_mount.format} table for {root_path} (general)")
                self._recurse_dir_if_needed(file_info, delta_log_folders)

        return delta_log_folders

    def _recurse_dir_if_needed(self, file_info: FileInfo, delta_log_folders: dict[str, TableInMount]) -> None:
        if self._is_recursible_dir(file_info):
            self._find_delta_log_folders(file_info.path, delta_log_folders)

    def _handle_partitioned_file(self, file_path, root_path, delta_log_folders):
        partition_format = self._find_partition_file_format(file_path)
        if partition_format:
            delta_log_folders[root_path] = partition_format
            logger.debug(f"Added {partition_format.format} table for {root_path} (partitioned)")

    def _find_partition_file_format(self, root_dir: str) -> TableInMount | None:
        logger.info(f"Listing partitioned file {root_dir}")
        file_infos = self._dbutils.fs.ls(root_dir)
        for file_info in file_infos:
            path_extension = self._assess_path(file_info)
            if path_extension:
                return TableInMount(format=path_extension.format, is_partitioned=True)
            if self._is_partitioned(file_info.name) and file_info.path != root_dir:
                return self._find_partition_file_format(file_info.path)
        return None

    def _assess_path(self, file_info: FileInfo) -> TableInMount | None:
        if file_info.name == "_delta_log/":
            logger.debug(f"Found delta table {file_info.path}")
            return TableInMount(format="DELTA", is_partitioned=False)
        if self._is_csv(file_info.name):
            logger.debug(f"Found csv {file_info.path}")
            return TableInMount(format="CSV", is_partitioned=False)
        if self._is_json(file_info.name):
            logger.debug(f"Found json {file_info.path}")
            return TableInMount(format="JSON", is_partitioned=False)
        if self._is_parquet(file_info.name):
            logger.debug(f"Found parquet {file_info.path}")
            return TableInMount(format="PARQUET", is_partitioned=False)
        return None

    def _is_recursible_dir(self, file_info: FileInfo) -> bool:
        # Rules for recursing into a folder
        # - should be size 0 (usually a directory)
        # - should not be the _delta_log directory
        # - file should not be partitioned
        # - file name should not start with 'part-'
        # - no brackets, brackets are not allowed in dbutils.fs.ls
        # - is not a streaming checkpoint dir
        return (
            file_info.size == 0
            and file_info.name != "_delta_log/"
            and not self._is_partitioned(file_info.name)
            and not file_info.name.startswith("part-")
            and not any(char in file_info.name for char in "[]")
            and not self._is_streaming_checkpoint_dir(file_info)
        )

    @staticmethod
    def _is_partitioned(file_name: str) -> bool:
        return '=' in file_name

    @staticmethod
    def _is_parquet(file_name: str) -> bool:
        parquet_patterns = {'.parquet'}
        return any(pattern in file_name for pattern in parquet_patterns)

    @staticmethod
    def _is_csv(file_name: str) -> bool:
        csv_patterns = {'.csv'}
        return any(pattern in file_name for pattern in csv_patterns)

    @staticmethod
    def _is_json(file_name: str) -> bool:
        json_patterns = {'.json'}
        return any(pattern in file_name for pattern in json_patterns)

    def _is_irrelevant(self, file_name: str) -> bool:
        return any(pattern in file_name for pattern in self._fiter_paths)

    def _is_streaming_checkpoint_dir(self, file_info: FileInfo) -> bool:
        path_dirs = [file.name for file in self._dbutils.fs.ls(file_info.path) if file.size == 0]
        return 'commits/' in path_dirs and 'offsets/' in path_dirs



from __future__ import annotations

import codecs
import dataclasses
import io
import logging
import sys
from abc import abstractmethod, ABC
from collections.abc import Iterable
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from typing import Any, BinaryIO, TextIO

from astroid import NodeNG  # type: ignore
from sqlglot import Expression, parse as parse_sql
from sqlglot.errors import SqlglotError

from databricks.sdk.service import compute
from databricks.sdk.service.workspace import Language

from databricks.labs.blueprint.paths import WorkspacePath


if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

# Code mapping between LSP, PyLint, and our own diagnostics:
# | LSP                       | PyLint     | Our            |
# |---------------------------|------------|----------------|
# | Severity.ERROR            | Error      | Failure()      |
# | Severity.WARN             | Warning    | Advisory()     |
# | DiagnosticTag.DEPRECATED  | Warning    | Deprecation()  |
# | Severity.INFO             | Info       | Advice()       |
# | Severity.HINT             | Convention | Convention()   |
# | DiagnosticTag.UNNECESSARY | Refactor   | Convention()   |

logger = logging.getLogger(__name__)


@dataclass
class Advice:
    code: str
    message: str
    # Lines and columns are both 0-based: the first line is line 0.
    start_line: int
    start_col: int
    end_line: int
    end_col: int

    def as_advisory(self) -> 'Advisory':
        return Advisory(**self.__dict__)

    def as_failure(self) -> 'Failure':
        return Failure(**self.__dict__)

    def as_deprecation(self) -> 'Deprecation':
        return Deprecation(**self.__dict__)

    def as_convention(self) -> 'Convention':
        return Convention(**self.__dict__)

    def for_path(self, path: Path) -> LocatedAdvice:
        return LocatedAdvice(self, path)

    @classmethod
    def from_node(cls, *, code: str, message: str, node: NodeNG) -> Advice:
        # Astroid lines are 1-based.
        return cls(
            code=code,
            message=message,
            start_line=(node.lineno or 1) - 1,
            start_col=node.col_offset or 0,
            end_line=(node.end_lineno or 1) - 1,
            end_col=node.end_col_offset,
        )

    def replace_from_node(self, node: NodeNG) -> Advice:
        # Astroid lines are 1-based.
        return dataclasses.replace(
            self,
            start_line=(node.lineno or 1) - 1,
            start_col=node.col_offset,
            end_line=(node.end_lineno or 1) - 1,
            end_col=node.end_col_offset,
        )


@dataclass
class LocatedAdvice:
    advice: Advice
    path: Path

    @property
    def is_unknown(self) -> bool:
        return self.path == Path('UNKNOWN')

    def message_relative_to(self, base: Path, *, default: Path | None = None) -> str:
        advice = self.advice
        path = self.path
        if self.is_unknown:
            logger.debug(f'THIS IS A BUG! {advice.code}:{advice.message} has unknown path')
        if default is not None:
            path = default
        try:
            path = path.relative_to(base)
        except ValueError:
            logger.debug(f'Not a relative path: {path} to base: {base}')
        # increment start_line because it is 0-based whereas IDEs are usually 1-based
        return f"./{path.as_posix()}:{advice.start_line+1}:{advice.start_col}: [{advice.code}] {advice.message}"


class Advisory(Advice):
    """A warning that does not prevent the code from running."""


class Failure(Advice):
    """An error that prevents the code from running."""


class Deprecation(Advice):
    """An advisory that suggests to replace the code with a newer version."""


class Convention(Advice):
    """A suggestion for a better way to write the code."""


class Linter:
    @abstractmethod
    def lint(self, code: str) -> Iterable[Advice]: ...


class SqlLinter(Linter):

    def lint(self, code: str) -> Iterable[Advice]:
        try:
            # TODO: unify with SqlParser.walk_expressions(...)
            expressions = parse_sql(code, read='databricks')
            for expression in expressions:
                if not expression:
                    continue
                yield from self.lint_expression(expression)
        except SqlglotError as e:
            logger.debug(f"Failed to parse SQL: {code}", exc_info=e)
            yield self.sql_parse_failure(code)

    @staticmethod
    def sql_parse_failure(code: str) -> Failure:
        return Failure(
            code='sql-parse-error',
            message=f"SQL expression is not supported yet: {code}",
            # SQLGlot does not propagate tokens yet. See https://github.com/tobymao/sqlglot/issues/3159
            start_line=0,
            start_col=0,
            end_line=0,
            end_col=1024,
        )

    @abstractmethod
    def lint_expression(self, expression: Expression) -> Iterable[Advice]: ...


class Fixer(ABC):

    @property
    @abstractmethod
    def name(self) -> str: ...

    @abstractmethod
    def apply(self, code: str) -> str: ...


@dataclass
class LineageAtom:

    object_type: str
    object_id: str
    other: dict[str, str] | None = None


@dataclass
class SourceInfo:

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Self:
        source_lineage = data.get("source_lineage", None)
        if isinstance(source_lineage, list) and len(source_lineage) > 0 and isinstance(source_lineage[0], dict):
            lineage_atoms = [LineageAtom(**lineage) for lineage in source_lineage]
            data["source_lineage"] = lineage_atoms
        return cls(**data)

    UNKNOWN = "unknown"

    source_id: str = UNKNOWN

    source_timestamp: datetime = field(default_factory=lambda: datetime.fromtimestamp(0), compare=False)
    """Unused attribute, kept for legacy reasons"""

    source_lineage: list[LineageAtom] = field(default_factory=list)

    assessment_start_timestamp: datetime = field(default_factory=lambda: datetime.fromtimestamp(0), compare=False)
    """Unused attribute, kept for legacy reasons"""

    assessment_end_timestamp: datetime = field(default_factory=lambda: datetime.fromtimestamp(0), compare=False)
    """Unused attribute, kept for legacy reasons"""

    def replace_source(
        self,
        source_id: str | None = None,
        source_lineage: list[LineageAtom] | None = None,
        source_timestamp: datetime | None = None,
    ) -> Self:
        return dataclasses.replace(
            self,
            source_id=source_id or self.source_id,
            source_timestamp=source_timestamp or self.source_timestamp,
            source_lineage=source_lineage or self.source_lineage,
        )

    def replace_assessment_infos(
        self,
        assessment_start: datetime | None = None,
        assessment_end: datetime | None = None,
    ) -> Self:
        return dataclasses.replace(
            self,
            assessment_start_timestamp=assessment_start or self.assessment_start_timestamp,
            assessment_end_timestamp=assessment_end or self.assessment_end_timestamp,
        )

    @property
    def source_type(self) -> str | None:
        if not self.source_lineage:
            return None
        last = self.source_lineage[-1]
        return last.object_type

    @property
    def query_id(self) -> str | None:
        if self.source_type != 'QUERY':
            return None
        last = self.source_lineage[-1]
        parts = last.object_id.split('/')
        if len(parts) < 2:
            return None
        return parts[1]


@dataclass
class UsedTable(SourceInfo):

    @classmethod
    def parse(cls, value: str, default_schema: str, is_read=True, is_write=False) -> UsedTable:
        parts = value.split(".")
        if len(parts) >= 3:
            catalog_name = parts.pop(0)
        else:
            catalog_name = "hive_metastore"
        if len(parts) >= 2:
            schema_name = parts.pop(0)
        else:
            schema_name = default_schema
        return UsedTable(
            catalog_name=catalog_name, schema_name=schema_name, table_name=parts[0], is_read=is_read, is_write=is_write
        )

    @property
    def full_name(self) -> str:
        return ".".join([self.catalog_name, self.schema_name, self.table_name])

    catalog_name: str = SourceInfo.UNKNOWN
    schema_name: str = SourceInfo.UNKNOWN
    table_name: str = SourceInfo.UNKNOWN
    is_read: bool = True
    is_write: bool = False


class TableCollector(ABC):

    @abstractmethod
    def collect_tables(self, source_code: str) -> Iterable[UsedTable]: ...


@dataclass
class UsedTableNode:
    table: UsedTable
    node: NodeNG


class TableSqlCollector(TableCollector, ABC): ...


@dataclass
class DirectFsAccess(SourceInfo):
    """A record describing a Direct File System Access"""

    path: str = SourceInfo.UNKNOWN
    is_read: bool = False
    is_write: bool = False


@dataclass
class DirectFsAccessNode:
    dfsa: DirectFsAccess
    node: NodeNG


class DfsaCollector(ABC):

    @abstractmethod
    def collect_dfsas(self, source_code: str) -> Iterable[DirectFsAccess]: ...


class DfsaSqlCollector(DfsaCollector, ABC): ...


# The default schema to use when the schema is not specified in a table reference
# See: https://spark.apache.org/docs/3.0.0-preview/sql-ref-syntax-qry-select-usedb.html
DEFAULT_CATALOG = 'hive_metastore'
DEFAULT_SCHEMA = 'default'


@dataclass
class CurrentSessionState:
    """
    A data class that represents the current state of a session.

    This class can be used to track various aspects of a session, such as the current schema.

    Attributes:
        catalog (str): The current schema of the session. If not provided, it defaults to 'DEFAULT_CATALOG'.
        schema (str): The current schema of the session. If not provided, it defaults to 'DEFAULT_SCHEMA'.
    """

    schema: str = DEFAULT_SCHEMA
    catalog: str = DEFAULT_CATALOG
    spark_conf: dict[str, str] | None = None
    named_parameters: dict[str, str] | None = None
    data_security_mode: compute.DataSecurityMode | None = None
    is_serverless: bool = False
    dbr_version: tuple[int, int] | None = None

    @classmethod
    def from_json(cls, json: dict) -> CurrentSessionState:
        return cls(
            schema=json.get('schema', DEFAULT_SCHEMA),
            catalog=json.get('catalog', DEFAULT_CATALOG),
            spark_conf=json.get('spark_conf', None),
            named_parameters=json.get('named_parameters', None),
            data_security_mode=cls.parse_security_mode(json.get('data_security_mode', None)),
            is_serverless=json.get('is_serverless', False),
            dbr_version=tuple(json['dbr_version']) if 'dbr_version' in json else None,
        )

    @staticmethod
    def parse_security_mode(mode_str: str | None) -> compute.DataSecurityMode | None:
        try:
            return compute.DataSecurityMode(mode_str) if mode_str else None
        except ValueError:
            logger.warning(f'Unknown data_security_mode {mode_str}')
            return None


class SqlSequentialLinter(SqlLinter, DfsaCollector, TableCollector):

    def __init__(
        self,
        linters: list[SqlLinter],
        dfsa_collectors: list[DfsaSqlCollector],
        table_collectors: list[TableSqlCollector],
    ):
        self._linters = linters
        self._dfsa_collectors = dfsa_collectors
        self._table_collectors = table_collectors

    def lint_expression(self, expression: Expression) -> Iterable[Advice]:
        for linter in self._linters:
            yield from linter.lint_expression(expression)

    def collect_dfsas(self, source_code: str) -> Iterable[DirectFsAccess]:
        for collector in self._dfsa_collectors:
            yield from collector.collect_dfsas(source_code)

    def collect_tables(self, source_code: str) -> Iterable[UsedTable]:
        for collector in self._table_collectors:
            yield from collector.collect_tables(source_code)


SUPPORTED_EXTENSION_LANGUAGES = {
    '.py': Language.PYTHON,
    '.sql': Language.SQL,
}


def file_language(path: Path) -> Language | None:
    return SUPPORTED_EXTENSION_LANGUAGES.get(path.suffix.lower())


def _detect_encoding_bom(binary_io: BinaryIO, *, preserve_position: bool) -> str | None:
    # Peek at the first (up to) 4 bytes, preserving the file position if requested.
    position = binary_io.tell() if preserve_position else None
    try:
        maybe_bom = binary_io.read(4)
    finally:
        if position is not None:
            binary_io.seek(position)
    # For these encodings, TextIOWrapper will skip over the BOM during decoding.
    if maybe_bom.startswith(codecs.BOM_UTF32_LE) or maybe_bom.startswith(codecs.BOM_UTF32_BE):
        return "utf-32"
    if maybe_bom.startswith(codecs.BOM_UTF16_LE) or maybe_bom.startswith(codecs.BOM_UTF16_BE):
        return "utf-16"
    if maybe_bom.startswith(codecs.BOM_UTF8):
        return "utf-8-sig"
    return None


def decode_with_bom(
    file: BinaryIO,
    encoding: str | None = None,
    errors: str | None = None,
    newline: str | None = None,
) -> TextIO:
    """Wrap an open binary file with a text decoder.

    This has the same semantics as the built-in `open()` call, except that if the encoding is not specified and the
    file is seekable then it will be checked for a BOM. If a BOM marker is found, that encoding is used. When neither
    an encoding nor a BOM are present the encoding of the system locale is used.

    Args:
          file: the open (binary) file to wrap in text mode.
          encoding: force decoding with a specific locale. If not present the file BOM and system locale are used.
          errors: how decoding errors should be handled, as per open().
          newline: how newlines should be handled, as per open().
    Raises:
          ValueError: if the encoding should be detected via potential BOM marker but the file is not seekable.
    Returns:
          a text-based IO wrapper that will decode the underlying binary-mode file as text.
    """
    use_encoding = _detect_encoding_bom(file, preserve_position=True) if encoding is None else encoding
    return io.TextIOWrapper(file, encoding=use_encoding, errors=errors, newline=newline)


def read_text(path: Path, size: int = -1) -> str:
    """Read a file as text, decoding according to the BOM marker if that is present.

    This differs to the normal `.read_text()` method on path which does not support BOM markers.

    Arguments:
        path: the path to read text from.
        size: how much text (measured in characters) to read. If negative, all text is read. Less may be read if the
            file is smaller than the specified size.
    Returns:
        The string content of the file, up to the specified size.
    """
    with path.open("rb") as binary_io:
        # If the open file is seekable, we can detect the BOM and decode without re-opening.
        if binary_io.seekable():
            with decode_with_bom(binary_io) as f:
                return f.read(size)
        encoding = _detect_encoding_bom(binary_io, preserve_position=False)
    # Otherwise having read the BOM there's no way to rewind so we need to re-open and read from that.
    with path.open("rt", encoding=encoding) as f:
        return f.read(size)


def safe_read_text(path: Path, size: int = -1) -> str | None:
    """Safe read a text file by handling reading exceptions, see :func:_read_text.

    Returns
        str : Content of file
        None : If error occurred during reading.
    """
    try:
        return read_text(path, size=size)
    except (FileNotFoundError, UnicodeDecodeError, PermissionError) as e:
        logger.warning(f"Could not read file: {path}", exc_info=e)
        return None


# duplicated from CellLanguage to prevent cyclic import
LANGUAGE_COMMENT_PREFIXES = {Language.PYTHON: '#', Language.SCALA: '//', Language.SQL: '--'}
NOTEBOOK_HEADER = "Databricks notebook source"


def is_a_notebook(path: Path, content: str | None = None) -> bool:
    if isinstance(path, WorkspacePath):
        return path.is_notebook()
    if not path.is_file():
        return False
    language = file_language(path)
    if not language:
        return False
    magic_header = f"{LANGUAGE_COMMENT_PREFIXES.get(language)} {NOTEBOOK_HEADER}"
    if content is not None:
        return content.startswith(magic_header)
    file_header = safe_read_text(path, size=len(magic_header))
    return file_header == magic_header



import logging
from collections.abc import Iterable
from dataclasses import dataclass
from functools import partial

from databricks.labs.blueprint.parallel import Threads

from databricks.labs.ucx.framework.crawlers import CrawlerBase
from databricks.labs.ucx.framework.utils import escape_sql_identifier
from databricks.labs.ucx.hive_metastore import TablesCrawler
from databricks.labs.ucx.hive_metastore.tables import FasterTableScanCrawler, Table

logger = logging.getLogger(__name__)


@dataclass
class TableSize:
    catalog: str
    database: str
    name: str
    size_in_bytes: int


class TableSizeCrawler(CrawlerBase[TableSize]):
    # TODO: Ensure TablesCrawler and FasterTableScanCrawler share a common interface.
    def __init__(self, tables_crawler: TablesCrawler | FasterTableScanCrawler) -> None:
        """
        Initializes a TablesSizeCrawler instance.

        Args:
            tables_crawler (TablesCrawler): The crawler to use to obtain the table inventory.
        """
        # pylint: disable-next=import-error,import-outside-toplevel
        from pyspark.sql.session import SparkSession  # type: ignore[import-not-found]

        super().__init__(
            tables_crawler._sql_backend,
            "hive_metastore",
            tables_crawler._schema,
            "table_size",
            TableSize,
        )
        self._tables_crawler = tables_crawler
        self._spark = SparkSession.builder.getOrCreate()

    def _crawl(self) -> Iterable[TableSize]:
        """Crawls and lists tables using table crawler
        Identifies DBFS root tables and calculates the size for these.
        """
        tasks = []
        for table in self._tables_crawler.snapshot():
            if not table.kind == "TABLE":
                continue
            if not table.is_dbfs_root:
                continue
            tasks.append(partial(self._safe_get_table_size, table))
        return Threads.strict('DBFS root table sizes', tasks)

    def _try_fetch(self) -> Iterable[TableSize]:
        """Tries to load table information from the database or throws TABLE_OR_VIEW_NOT_FOUND error"""
        for row in self._fetch(f"SELECT * FROM {escape_sql_identifier(self.full_name)}"):
            yield TableSize(*row)

    def _safe_get_table_size(self, table: Table) -> TableSize | None:
        logger.debug(f"Evaluating {table.key} table size.")
        try:
            # refresh table statistics to avoid stale stats in HMS
            self._sql_backend.execute(f"ANALYZE table {table.safe_sql_key} compute STATISTICS NOSCAN")
            jvm_df = self._spark._jsparkSession.table(table.safe_sql_key)  # pylint: disable=protected-access
            size_in_bytes = jvm_df.queryExecution().analyzed().stats().sizeInBytes()
            return TableSize(
                catalog=table.catalog,
                database=table.database,
                name=table.name,
                size_in_bytes=size_in_bytes,
            )
        except Exception as e:  # pylint: disable=broad-exception-caught
            if "[TABLE_OR_VIEW_NOT_FOUND]" in str(e) or "[DELTA_TABLE_NOT_FOUND]" in str(e):
                logger.warning(f"Failed to evaluate {table.key} table size. Table not found.")
                return None
            if "[DELTA_INVALID_FORMAT]" in str(e):
                logger.warning(f"Unable to read Delta table {table.key}, please check table structure and try again.")
                return None
            if "[DELTA_MISSING_TRANSACTION_LOG]" in str(e):
                logger.warning(f"Delta table {table.key} is corrupt: missing transaction log.")
                return None
            logger.error(f"Failed to evaluate {table.key} table size: ", exc_info=True)

        return None



import logging
from collections.abc import Callable, Iterable
from typing import TypeVar

from sqlglot import parse
from sqlglot.errors import SqlglotError
from sqlglot.expressions import Create, Delete, Drop, Expression, Select, Table, Update, Use

from databricks.labs.ucx.source_code.base import UsedTable, CurrentSessionState

logger = logging.getLogger(__name__)

T = TypeVar("T")
E = TypeVar("E", bound=Expression)


class SqlExpression:

    def __init__(self, expression: Expression):
        self._expression = expression

    def collect_used_tables(self, required_catalog: str, session_state: CurrentSessionState) -> Iterable[UsedTable]:
        for table in self.find_all(Table):
            info = self._collect_table_info(table, required_catalog, session_state)
            if info:
                yield info

    def _collect_table_info(
        self,
        table: Table,
        required_catalog: str,
        session_state: CurrentSessionState,
    ) -> UsedTable | None:
        if isinstance(self._expression, Use):
            # Sqlglot captures the database name in the Use statement as a Table, with
            # the schema  as the table name.
            session_state.schema = table.name
            return None
        if isinstance(self._expression, Drop) and getattr(self._expression, "kind", None) == "SCHEMA":
            # Sqlglot captures the schema name in the Drop statement as a Table, with
            # the schema  as the db name.
            return None
        if isinstance(self._expression, Create) and getattr(self._expression, "kind", None) == "SCHEMA":
            # Sqlglot captures the schema name in the Create statement as a Table, with
            # the schema  as the db name.
            session_state.schema = table.db
            return None
        # we only return tables in the required catalog
        catalog_name = table.catalog or required_catalog
        if catalog_name != required_catalog:
            return None
        # Sqlglot uses db instead of schema, watch out for that
        src_schema = table.db if table.db else session_state.schema
        if not src_schema:
            logger.warning(f"Could not determine schema for table {table.name}")
            return None
        return UsedTable(
            catalog_name=catalog_name,
            schema_name=src_schema,
            table_name=table.name,
            is_read=isinstance(self._expression, Select),
            is_write=isinstance(self._expression, (Create, Update, Delete)),
        )

    def find_all(self, klass: type[E]) -> Iterable[E]:
        try:
            return self._expression.find_all(klass)
        except SqlglotError as e:
            logger.warning(f"Failed to find all {klass} in expression: {self._expression}", exc_info=e)
            return []


class SqlParser:

    @classmethod
    def walk_expressions(cls, sql_code: str, callback: Callable[[SqlExpression], Iterable[T]]) -> Iterable[T]:
        try:
            expressions = parse(sql_code, read='databricks')
            for expression in expressions:
                if not expression:
                    continue
                yield from callback(SqlExpression(expression))
        except SqlglotError as e:
            logger.warning(f"Failed to parse SQL: {sql_code}", exc_info=e)



import datetime
import logging
from dataclasses import dataclass, replace
from collections.abc import Iterable, KeysView
from typing import ClassVar

from databricks.labs.lsql.backends import SqlBackend
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import DatabricksError, NotFound
from databricks.sdk.service.catalog import CatalogInfo, CatalogInfoSecurableKind, SchemaInfo

from databricks.labs.ucx.framework.crawlers import CrawlerBase
from databricks.labs.ucx.framework.utils import escape_sql_identifier
from databricks.labs.ucx.hive_metastore.tables import TablesCrawler

logger = logging.getLogger(__name__)


@dataclass
class TableMigrationStatus:
    src_schema: str
    src_table: str
    dst_catalog: str | None = None
    dst_schema: str | None = None
    dst_table: str | None = None
    update_ts: str | None = None

    __id_attributes__: ClassVar[tuple[str, ...]] = ("src_schema", "src_table")

    def destination(self):
        return f"{self.dst_catalog}.{self.dst_schema}.{self.dst_table}".lower()

    @classmethod
    def from_json(cls, raw: dict[str, str]) -> "TableMigrationStatus":
        return cls(
            src_schema=raw['src_schema'],
            src_table=raw['src_table'],
            dst_catalog=raw.get('dst_catalog', None),
            dst_schema=raw.get('dst_schema', None),
            dst_table=raw.get('dst_table', None),
            update_ts=raw.get('update_ts', None),
        )


@dataclass(frozen=True)
class TableView:
    catalog: str
    schema: str
    name: str

    @property
    def key(self):
        return f"{self.catalog}.{self.schema}.{self.name}".lower()


class TableMigrationIndex:
    def __init__(self, tables: Iterable[TableMigrationStatus]):
        self._index = {(ms.src_schema, ms.src_table): ms for ms in tables}

    def is_migrated(self, schema: str, table: str) -> bool:
        """Check if a table is migrated."""
        return self.get(schema, table) is not None

    def get(self, schema: str, table: str) -> TableMigrationStatus | None:
        """Get the migration status for a table. If the table is not migrated, return None."""
        dst = self._index.get((schema.lower(), table.lower()))
        if not dst or not dst.dst_table:
            return None
        return dst

    def snapshot(self) -> KeysView[tuple[str, str]]:
        return self._index.keys()


class TableMigrationStatusRefresher(CrawlerBase[TableMigrationStatus]):
    """Crawler to capture the migration status of tables (and views).

    Migrated tables have a property set to mark them as such; this crawler scans all tables and views, examining the
    properties for the presence of the marker.
    """

    _skip_catalog_securable_kinds = [
        CatalogInfoSecurableKind.CATALOG_INTERNAL,
        CatalogInfoSecurableKind.CATALOG_SYSTEM,
    ]

    def __init__(self, ws: WorkspaceClient, sql_backend: SqlBackend, schema, tables_crawler: TablesCrawler):
        super().__init__(sql_backend, "hive_metastore", schema, "migration_status", TableMigrationStatus)
        self._ws = ws
        self._tables_crawler = tables_crawler

    def index(self, *, force_refresh: bool = False) -> TableMigrationIndex:
        return TableMigrationIndex(list(self.snapshot(force_refresh=force_refresh)))

    def get_seen_tables(self) -> dict[str, str]:
        seen_tables: dict[str, str] = {}
        for schema in self._iter_schemas():
            if schema.catalog_name is None or schema.name is None:
                continue
            try:
                # ws.tables.list returns Iterator[TableInfo], so we need to convert it to a list in order to catch the exception
                tables = list(self._ws.tables.list(catalog_name=schema.catalog_name, schema_name=schema.name))
            except NotFound:
                logger.warning(f"Schema {schema.full_name} no longer exists. Skipping checking its migration status.")
                continue
            except DatabricksError as e:
                logger.warning(f"Error while listing tables in schema: {schema.full_name}", exc_info=e)
                continue
            for table in tables:
                if not table.properties:
                    continue
                if "upgraded_from" not in table.properties:
                    continue
                if not table.full_name:
                    logger.warning(f"The table {table.name} in {schema.name} has no full name")
                    continue
                seen_tables[table.full_name.lower()] = table.properties["upgraded_from"].lower()
        return seen_tables

    def is_migrated(self, schema: str, table: str) -> bool:
        try:
            results = self._sql_backend.fetch(
                f"SHOW TBLPROPERTIES {escape_sql_identifier(schema + '.' + table)} ('upgraded_to')"
            )
            for result in results:
                if "does not have property" in result.value:
                    continue
                logger.info(f"{schema}.{table} is set as migrated")
                return True
            logger.info(f"{schema}.{table} is set as not migrated")
        except NotFound:
            # If the source table doesn't exist anymore, we mark it as migrated to avoid trying to migrate it again and
            # for views that have the table as dependency
            logger.warning(f"failed-to-migrate: {schema}.{table} set as a source does no longer exist")
            return True
        return False

    def _crawl(self) -> Iterable[TableMigrationStatus]:
        all_tables = self._tables_crawler.snapshot()
        reverse_seen = {v: k for k, v in self.get_seen_tables().items()}
        timestamp = datetime.datetime.now(datetime.timezone.utc).timestamp()
        for table in all_tables:
            src_schema = table.database.lower()
            src_table = table.name.lower()
            table_migration_status = TableMigrationStatus(
                src_schema=src_schema, src_table=src_table, update_ts=str(timestamp)
            )
            if table.key in reverse_seen and self.is_migrated(src_schema, src_table):
                target_table = reverse_seen[table.key]
                if len(target_table.split(".")) == 3:
                    table_migration_status = replace(
                        table_migration_status,
                        dst_catalog=target_table.split(".")[0],
                        dst_schema=target_table.split(".")[1],
                        dst_table=target_table.split(".")[2],
                    )
            yield table_migration_status

    def _try_fetch(self) -> Iterable[TableMigrationStatus]:
        for row in self._fetch(f"SELECT * FROM {escape_sql_identifier(self.full_name)}"):
            yield TableMigrationStatus(*row)

    def _iter_catalogs(self) -> Iterable[CatalogInfo]:
        try:
            for catalog in self._ws.catalogs.list():
                if catalog.securable_kind in self._skip_catalog_securable_kinds:
                    continue
                yield catalog
        except DatabricksError as e:
            logger.error("Cannot list catalogs", exc_info=e)

    def _iter_schemas(self) -> Iterable[SchemaInfo]:
        for catalog in self._iter_catalogs():
            if catalog.name is None:
                continue
            try:
                yield from self._ws.schemas.list(catalog_name=catalog.name)
            except NotFound:
                logger.warning(f"Catalog {catalog.name} no longer exists. Skipping checking its migration status.")
                continue
            except DatabricksError as e:
                logger.warning(f"Error while listing schemas in catalog: {catalog.name}", exc_info=e)
                continue



from __future__ import annotations

import builtins
import sys
from abc import ABC, abstractmethod
import logging
import re
from collections.abc import Iterable
from dataclasses import dataclass
from typing import TypeVar, cast

from astroid import (  # type: ignore
    Assign,
    AssignName,
    Attribute,
    Call,
    ClassDef,
    Const,
    Expr,
    Import,
    ImportFrom,
    Instance,
    Module,
    Name,
    NodeNG,
    parse,
    Uninferable,
)

from databricks.labs.ucx.source_code.base import (
    Failure,
    Linter,
    Advice,
    TableCollector,
    UsedTable,
    UsedTableNode,
    DfsaCollector,
    DirectFsAccess,
    DirectFsAccessNode,
)

logger = logging.getLogger(__name__)
missing_handlers: set[str] = set()


T = TypeVar("T", bound=NodeNG)


@dataclass(frozen=True)
class MaybeTree:
    tree: Tree | None
    failure: Failure | None

    def walk(self) -> Iterable[NodeNG]:
        # mainly a helper method for unit testing
        if self.tree is None:  # no cov
            assert self.failure is not None
            logger.warning(self.failure.message)
            return []
        return self.tree.walk()

    def first_statement(self) -> NodeNG | None:
        # mainly a helper method for unit testing
        if self.tree is None:  # no cov
            assert self.failure is not None
            logger.warning(self.failure.message)
            return None
        return self.tree.first_statement()


class Tree:  # pylint: disable=too-many-public-methods

    @classmethod
    def maybe_parse(cls, code: str) -> MaybeTree:
        try:
            root = parse(code)
            tree = Tree(root)
            return MaybeTree(tree, None)
        except Exception as e:  # pylint: disable=broad-exception-caught
            # see https://github.com/databrickslabs/ucx/issues/2976
            return cls._definitely_failure('system-error', code, e)

    @staticmethod
    def _definitely_failure(message_code: str, source_code: str, e: Exception) -> MaybeTree:
        return MaybeTree(
            None,
            Failure(
                code=message_code,
                message=f"Failed to parse code `{source_code}`: {type(e)}: {e}. Report this as an issue on UCX GitHub.",
                # Lines and columns are both 0-based: the first line is line 0.
                start_line=0,
                start_col=0,
                end_line=1,
                end_col=1,
            ),
        )

    @classmethod
    def maybe_normalized_parse(cls, code: str) -> MaybeTree:
        code = cls.normalize(code)
        return cls.maybe_parse(code)

    @classmethod
    def normalize(cls, code: str) -> str:
        code = cls._normalize_indents(code)
        code = cls._convert_magic_lines_to_magic_commands(code)
        return code

    @classmethod
    def _normalize_indents(cls, python_code: str) -> str:
        lines = python_code.split("\n")
        for line in lines:
            # skip leading ws and comments
            if len(line.strip()) == 0 or line.startswith('#'):
                continue
            if not line.startswith(' '):
                # first line of code is correctly indented
                return python_code
            # first line of code is indented when it shouldn't
            prefix_count = len(line) - len(line.lstrip(' '))
            prefix_str = ' ' * prefix_count
            for i, line_to_fix in enumerate(lines):
                if line_to_fix.startswith(prefix_str):
                    lines[i] = line_to_fix[prefix_count:]
            return "\n".join(lines)
        return python_code

    @classmethod
    def _convert_magic_lines_to_magic_commands(cls, python_code: str) -> str:
        lines = python_code.split("\n")
        magic_markers = {"%", "!"}
        in_multi_line_comment = False
        pattern = re.compile('"""')
        for i, line in enumerate(lines):
            if len(line) == 0:
                continue
            if not in_multi_line_comment and line[0] in magic_markers:
                lines[i] = f"magic_command({line.encode()!r})"
                continue
            matches = re.findall(pattern, line)
            if len(matches) & 1:
                in_multi_line_comment = not in_multi_line_comment
        return "\n".join(lines)

    @classmethod
    def new_module(cls) -> Tree:
        node = Module("root")
        return Tree(node)

    def __init__(self, node: NodeNG):
        self._node: NodeNG = node

    @property
    def node(self) -> NodeNG:
        return self._node

    @property
    def root(self) -> NodeNG:
        node = self._node
        while node.parent:
            node = node.parent
        return node

    def walk(self) -> Iterable[NodeNG]:
        yield from self._walk(self._node)

    @classmethod
    def _walk(cls, node: NodeNG) -> Iterable[NodeNG]:
        yield node
        for child in node.get_children():
            yield from cls._walk(child)

    def locate(self, node_type: type[T], match_nodes: list[tuple[str, type]]) -> list[T]:
        visitor = MatchingVisitor(node_type, match_nodes)
        visitor.visit(self._node)
        return visitor.matched_nodes

    def first_statement(self) -> NodeNG | None:
        if isinstance(self._node, Module):
            if len(self._node.body) > 0:
                return self._node.body[0]
        return None

    def __repr__(self):
        truncate_after = 32
        code = repr(self._node)
        if len(code) > truncate_after:
            code = code[0:truncate_after] + "..."
        return f"<Tree: {code}>"

    def append_tree(self, tree: Tree) -> Tree:
        """returns the appended tree, not the consolidated one!"""
        if not isinstance(tree.node, Module):
            raise NotImplementedError(f"Can't append tree from {type(tree.node).__name__}")
        tree_module: Module = cast(Module, tree.node)
        self.append_nodes(tree_module.body)
        self.append_globals(tree_module.globals)
        # the following may seem strange but it's actually ok to use the original module as tree root
        # because each node points to the correct parent (practically, the tree is now only a list of statements)
        return tree

    def append_globals(self, globs: dict[str, list[Expr]]) -> None:
        if not isinstance(self.node, Module):
            raise NotImplementedError(f"Can't append globals to {type(self.node).__name__}")
        self_module: Module = cast(Module, self.node)
        for name, values in globs.items():
            statements: list[Expr] = self_module.globals.get(name, None)
            if statements is None:
                self_module.globals[name] = list(values)  # clone the source list to avoid side-effects
                continue
            statements.extend(values)

    def append_nodes(self, nodes: list[NodeNG]) -> None:
        if not isinstance(self.node, Module):
            raise NotImplementedError(f"Can't append statements to {type(self.node).__name__}")
        self_module: Module = cast(Module, self.node)
        for node in nodes:
            node.parent = self_module
            self_module.body.append(node)

    def is_instance_of(self, class_name: str) -> bool:
        for inferred in self.node.inferred():
            if inferred is Uninferable:
                continue
            if not isinstance(inferred, (Const, Instance)):
                return False
            proxied = getattr(inferred, "_proxied", None)
            return isinstance(proxied, ClassDef) and proxied.name == class_name
        return False

    def is_from_module(self, module_name: str) -> bool:
        return self._is_from_module(module_name, set())

    def _is_from_module(self, module_name: str, visited: set[NodeNG]) -> bool:
        if self._node in visited:
            logger.debug(f"Recursion encountered while traversing node {self._node.as_string()}")
            return False
        visited.add(self._node)
        return self._node_is_from_module(module_name, visited)

    def _node_is_from_module(self, module_name: str, visited: set[NodeNG]) -> bool:
        if isinstance(self._node, Name):
            return self._name_is_from_module(module_name, visited)
        if isinstance(self._node, Call):
            return self._call_is_from_module(module_name, visited)
        if isinstance(self._node, Attribute):
            return self._attribute_is_from_module(module_name, visited)
        if isinstance(self._node, Const):
            return self._const_is_from_module(module_name, visited)
        return False

    def _name_is_from_module(self, module_name: str, visited: set[NodeNG]) -> bool:
        assert isinstance(self._node, Name)
        # if this is the call's root node, check it against the required module
        if self._node.name == module_name:
            return True
        root = self.root
        if not isinstance(root, Module):
            return False
        for value in root.globals.get(self._node.name, []):
            if not isinstance(value, AssignName) or not isinstance(value.parent, Assign):
                continue
            if _LocalTree(value.parent.value).is_from_module_visited(module_name, visited):
                return True
        return False

    def _call_is_from_module(self, module_name: str, visited: set[NodeNG]) -> bool:
        assert isinstance(self._node, Call)
        # walk up intermediate calls such as spark.range(...)
        return isinstance(self._node.func, Attribute) and _LocalTree(self._node.func.expr).is_from_module_visited(
            module_name, visited
        )

    def _attribute_is_from_module(self, module_name: str, visited: set[NodeNG]) -> bool:
        assert isinstance(self._node, Attribute)
        return _LocalTree(self._node.expr).is_from_module_visited(module_name, visited)

    def _const_is_from_module(self, module_name: str, visited: set[NodeNG]) -> bool:
        assert isinstance(self._node, Const)
        return _LocalTree(self._node.parent).is_from_module_visited(module_name, visited)

    def has_global(self, name: str) -> bool:
        if not isinstance(self.node, Module):
            return False
        self_module: Module = cast(Module, self.node)
        return self_module.globals.get(name, None) is not None

    def get_global(self, name: str) -> list[NodeNG]:
        if not self.has_global(name):
            return []
        return cast(Module, self.node).globals.get(name)

    def nodes_between(self, first_line: int, last_line: int) -> list[NodeNG]:
        if not isinstance(self.node, Module):
            raise NotImplementedError(f"Can't extract nodes from {type(self.node).__name__}")
        self_module: Module = cast(Module, self.node)
        nodes: list[NodeNG] = []
        for node in self_module.body:
            if node.lineno < first_line:
                continue
            if node.lineno > last_line:
                break
            nodes.append(node)
        return nodes

    def globals_between(self, first_line: int, last_line: int) -> dict[str, list[NodeNG]]:
        if not isinstance(self.node, Module):
            raise NotImplementedError(f"Can't extract globals from {type(self.node).__name__}")
        self_module: Module = cast(Module, self.node)
        globs: dict[str, list[NodeNG]] = {}
        for key, nodes in self_module.globals.items():
            nodes_in_scope: list[NodeNG] = []
            for node in nodes:
                if node.lineno < first_line or node.lineno > last_line:
                    continue
                nodes_in_scope.append(node)
            if len(nodes_in_scope) > 0:
                globs[key] = nodes_in_scope
        return globs

    def line_count(self) -> int:
        if not isinstance(self.node, Module):
            raise NotImplementedError(f"Can't count lines from {type(self.node).__name__}")
        self_module: Module = cast(Module, self.node)
        nodes_count = len(self_module.body)
        if nodes_count == 0:
            return 0
        return 1 + self_module.body[nodes_count - 1].lineno - self_module.body[0].lineno

    def renumber(self, start: int) -> Tree:
        assert start != 0
        if not isinstance(self.node, Module):
            raise NotImplementedError(f"Can't renumber {type(self.node).__name__}")
        root: Module = cast(Module, self.node)
        # for now renumber in place to avoid the complexity of rebuilding the tree with clones

        def renumber_node(node: NodeNG, offset: int) -> None:
            for child in node.get_children():
                renumber_node(child, offset + (child.lineno or 0) - (node.lineno or 0))
            if node.end_lineno:
                node.end_lineno = node.end_lineno + offset
                node.lineno = node.lineno + offset

        nodes = root.body if start > 0 else reversed(root.body)
        for node in nodes:
            offset = start - node.lineno
            renumber_node(node, offset)
            num_lines = 1 + (node.end_lineno - node.lineno if node.end_lineno else 0)
            start = start + num_lines if start > 0 else start - num_lines
        return self

    def is_builtin(self) -> bool:
        if isinstance(self._node, Name):
            name = self._node.name
            return name in dir(builtins) or name in sys.stdlib_module_names or name in sys.builtin_module_names
        if isinstance(self._node, Call):
            return Tree(self._node.func).is_builtin()
        if isinstance(self._node, Attribute):
            return Tree(self._node.expr).is_builtin()
        return False  # not supported yet


class _LocalTree(Tree):

    def is_from_module_visited(self, name: str, visited_nodes: set[NodeNG]) -> bool:
        return self._is_from_module(name, visited_nodes)


class TreeHelper(ABC):

    @classmethod
    def get_call_name(cls, call: Call) -> str:
        if not isinstance(call, Call):
            return ""
        func = call.func
        if isinstance(func, Name):
            return func.name
        if isinstance(func, Attribute):
            return func.attrname
        return ""  # not supported yet

    @classmethod
    def extract_call_by_name(cls, call: Call, name: str) -> Call | None:
        """Given a call-chain, extract its sub-call by method name (if it has one)"""
        assert isinstance(call, Call)
        node = call
        while True:
            func = node.func
            if not isinstance(func, Attribute):
                return None
            if func.attrname == name:
                return node
            if not isinstance(func.expr, Call):
                return None
            node = func.expr

    @classmethod
    def args_count(cls, node: Call) -> int:
        """Count the number of arguments (positionals + keywords)"""
        assert isinstance(node, Call)
        return len(node.args) + len(node.keywords)

    @classmethod
    def get_arg(
        cls,
        node: Call,
        arg_index: int | None,
        arg_name: str | None,
    ) -> NodeNG | None:
        """Extract the call argument identified by an optional position or name (if it has one)"""
        assert isinstance(node, Call)
        if arg_index is not None and len(node.args) > arg_index:
            return node.args[arg_index]
        if arg_name is not None:
            arg = [kw.value for kw in node.keywords if kw.arg == arg_name]
            if len(arg) == 1:
                return arg[0]
        return None

    @classmethod
    def is_none(cls, node: NodeNG) -> bool:
        """Check if the given AST expression is the None constant"""
        if not isinstance(node, Const):
            return False
        return node.value is None

    @classmethod
    def get_full_attribute_name(cls, node: Attribute) -> str | None:
        return cls._get_attribute_value(node)

    @classmethod
    def get_function_name(cls, node: Call) -> str | None:
        if not isinstance(node, Call):
            return None
        if isinstance(node.func, Attribute):
            return node.func.attrname
        if isinstance(node.func, Name):
            return node.func.name
        return None

    @classmethod
    def get_full_function_name(cls, node: Call) -> str | None:
        if not isinstance(node, Call):
            return None
        if isinstance(node.func, Attribute):
            return cls._get_attribute_value(node.func)
        if isinstance(node.func, Name):
            return node.func.name
        return None

    @classmethod
    def _get_attribute_value(cls, node: Attribute) -> str | None:
        if isinstance(node.expr, Name):
            return node.expr.name + '.' + node.attrname
        if isinstance(node.expr, Attribute):
            parent = cls._get_attribute_value(node.expr)
            return node.attrname if parent is None else parent + '.' + node.attrname
        if isinstance(node.expr, Call):
            name = cls.get_full_function_name(node.expr)
            return node.attrname if name is None else name + '.' + node.attrname
        name = type(node.expr).__name__
        if name not in missing_handlers:
            missing_handlers.add(name)
            logger.debug(f"Missing handler for {name}")
        return None


class TreeVisitor:

    def visit(self, node: NodeNG) -> None:
        self._visit_specific(node)
        for child in node.get_children():
            self.visit(child)

    def _visit_specific(self, node: NodeNG) -> None:
        method_name = "visit_" + type(node).__name__.lower()
        method_slot = getattr(self, method_name, None)
        if callable(method_slot):
            method_slot(node)
            return
        self.visit_nodeng(node)

    def visit_nodeng(self, node: NodeNG) -> None:
        pass


class MatchingVisitor(TreeVisitor):

    def __init__(self, node_type: type, match_nodes: list[tuple[str, type]]):
        super()
        self._matched_nodes: list[NodeNG] = []
        self._node_type = node_type
        self._match_nodes = match_nodes
        self._imports: dict[str, list[NodeNG]] = {}

    @property
    def matched_nodes(self) -> list[NodeNG]:
        return self._matched_nodes

    def visit_assign(self, node: Assign) -> None:
        if self._node_type is not Assign:
            return
        self._matched_nodes.append(node)

    def visit_call(self, node: Call) -> None:
        if self._node_type is not Call:
            return
        try:
            if self._matches(node.func, 0):
                self._matched_nodes.append(node)
        except NotImplementedError as e:
            logger.warning(f"Missing implementation: {e.args[0]}")

    def visit_import(self, node: Import) -> None:
        if self._node_type is not Import:
            return
        self._matched_nodes.append(node)

    def visit_importfrom(self, node: ImportFrom) -> None:
        if self._node_type is not ImportFrom:
            return
        self._matched_nodes.append(node)

    def _matches(self, node: NodeNG, depth: int) -> bool:
        if depth >= len(self._match_nodes):
            return False
        if isinstance(node, Call):
            return self._matches(node.func, depth)
        name, match_node = self._match_nodes[depth]
        node = self._adjust_node_for_import_member(name, match_node, node)
        if not isinstance(node, match_node):
            return False
        next_node: NodeNG | None = None
        if isinstance(node, Attribute):
            if node.attrname != name:
                return False
            next_node = node.expr
        elif isinstance(node, Name):
            if node.name != name:
                return False
        else:
            raise NotImplementedError(str(type(node)))
        if next_node is None:
            # is this the last node to match ?
            return len(self._match_nodes) - 1 == depth
        return self._matches(next_node, depth + 1)

    def _adjust_node_for_import_member(self, name: str, match_node: type, node: NodeNG) -> NodeNG:
        if isinstance(node, match_node):
            return node
        # if we're looking for an attribute, it might be a global name
        if match_node != Attribute or not isinstance(node, Name) or node.name != name:
            return node
        # in which case it could be an import member
        module = Tree(Tree(node).root)
        if not module.has_global(node.name):
            return node
        for import_from in module.get_global(node.name):
            if not isinstance(import_from, ImportFrom):
                continue
            parent = Name(
                name=import_from.modname,
                lineno=import_from.lineno,
                col_offset=import_from.col_offset,
                end_lineno=import_from.end_lineno,
                end_col_offset=import_from.end_col_offset,
                parent=import_from.parent,
            )
            resolved = Attribute(
                attrname=name,
                lineno=import_from.lineno,
                col_offset=import_from.col_offset,
                end_lineno=import_from.end_lineno,
                end_col_offset=import_from.end_col_offset,
                parent=parent,
            )
            resolved.postinit(parent)
            return resolved
        return node


class NodeBase(ABC):

    def __init__(self, node: NodeNG):
        self._node = node

    @property
    def node(self) -> NodeNG:
        return self._node

    def __repr__(self):
        return f"<{self.__class__.__name__}: {repr(self._node)}>"


class PythonLinter(Linter):

    def lint(self, code: str) -> Iterable[Advice]:
        maybe_tree = Tree.maybe_normalized_parse(code)
        if maybe_tree.failure:
            yield maybe_tree.failure
            return
        assert maybe_tree.tree is not None
        yield from self.lint_tree(maybe_tree.tree)

    @abstractmethod
    def lint_tree(self, tree: Tree) -> Iterable[Advice]: ...


class TablePyCollector(TableCollector, ABC):

    def collect_tables(self, source_code: str) -> Iterable[UsedTable]:
        maybe_tree = Tree.maybe_normalized_parse(source_code)
        if maybe_tree.failure:
            logger.warning(maybe_tree.failure.message)
            return
        assert maybe_tree.tree is not None
        for table_node in self.collect_tables_from_tree(maybe_tree.tree):
            yield table_node.table

    @abstractmethod
    def collect_tables_from_tree(self, tree: Tree) -> Iterable[UsedTableNode]: ...


class DfsaPyCollector(DfsaCollector, ABC):

    def collect_dfsas(self, source_code: str) -> Iterable[DirectFsAccess]:
        maybe_tree = Tree.maybe_normalized_parse(source_code)
        if maybe_tree.failure:
            logger.warning(maybe_tree.failure.message)
            return
        assert maybe_tree.tree is not None
        for dfsa_node in self.collect_dfsas_from_tree(maybe_tree.tree):
            yield dfsa_node.dfsa

    @abstractmethod
    def collect_dfsas_from_tree(self, tree: Tree) -> Iterable[DirectFsAccessNode]: ...


class PythonSequentialLinter(Linter, DfsaCollector, TableCollector):

    def __init__(
        self,
        linters: list[PythonLinter],
        dfsa_collectors: list[DfsaPyCollector],
        table_collectors: list[TablePyCollector],
    ):
        self._linters = linters
        self._dfsa_collectors = dfsa_collectors
        self._table_collectors = table_collectors
        self._tree: Tree | None = None

    def lint(self, code: str) -> Iterable[Advice]:
        maybe_tree = self._parse_and_append(code)
        if maybe_tree.failure:
            yield maybe_tree.failure
            return
        assert maybe_tree.tree is not None
        yield from self.lint_tree(maybe_tree.tree)

    def lint_tree(self, tree: Tree) -> Iterable[Advice]:
        for linter in self._linters:
            yield from linter.lint_tree(tree)

    def _parse_and_append(self, code: str) -> MaybeTree:
        maybe_tree = Tree.maybe_normalized_parse(code)
        if maybe_tree.failure:
            return maybe_tree
        assert maybe_tree.tree is not None
        self.append_tree(maybe_tree.tree)
        return maybe_tree

    def append_tree(self, tree: Tree) -> None:
        self._make_tree().append_tree(tree)

    def append_nodes(self, nodes: list[NodeNG]) -> None:
        self._make_tree().append_nodes(nodes)

    def append_globals(self, globs: dict) -> None:
        self._make_tree().append_globals(globs)

    def process_child_cell(self, code: str) -> None:
        this_tree = self._make_tree()
        maybe_tree = Tree.maybe_normalized_parse(code)
        if maybe_tree.failure:
            # TODO: bubble up this error
            logger.warning(maybe_tree.failure.message)
            return
        assert maybe_tree.tree is not None
        this_tree.append_tree(maybe_tree.tree)

    def collect_dfsas(self, source_code: str) -> Iterable[DirectFsAccess]:
        maybe_tree = self._parse_and_append(source_code)
        if maybe_tree.failure:
            logger.warning(maybe_tree.failure.message)
            return
        assert maybe_tree.tree is not None
        for dfsa_node in self.collect_dfsas_from_tree(maybe_tree.tree):
            yield dfsa_node.dfsa

    def collect_dfsas_from_tree(self, tree: Tree) -> Iterable[DirectFsAccessNode]:
        for collector in self._dfsa_collectors:
            yield from collector.collect_dfsas_from_tree(tree)

    def collect_tables(self, source_code: str) -> Iterable[UsedTable]:
        maybe_tree = self._parse_and_append(source_code)
        if maybe_tree.failure:
            logger.warning(maybe_tree.failure.message)
            return
        assert maybe_tree.tree is not None
        for table_node in self.collect_tables_from_tree(maybe_tree.tree):
            yield table_node.table

    def collect_tables_from_tree(self, tree: Tree) -> Iterable[UsedTableNode]:
        for collector in self._table_collectors:
            yield from collector.collect_tables_from_tree(tree)

    def _make_tree(self) -> Tree:
        if self._tree is None:
            self._tree = Tree.new_module()
        return self._tree



import logging
import re
import typing
from collections.abc import Collection, Iterable, Iterator
from dataclasses import dataclass
from enum import Enum, auto
from functools import cached_property, partial

import sqlglot
from sqlglot import expressions
from sqlglot.expressions import LocationProperty
from sqlglot.errors import ParseError

from databricks.labs.blueprint.parallel import Threads
from databricks.labs.lsql.backends import SqlBackend
from databricks.sdk.errors import NotFound

from databricks.labs.ucx.framework.crawlers import CrawlerBase
from databricks.labs.ucx.framework.utils import escape_sql_identifier

logger = logging.getLogger(__name__)


class What(Enum):
    EXTERNAL_SYNC = auto()
    EXTERNAL_HIVESERDE = auto()
    EXTERNAL_NO_SYNC = auto()
    DBFS_ROOT_DELTA = auto()
    DBFS_ROOT_NON_DELTA = auto()
    VIEW = auto()
    TABLE_IN_MOUNT = auto()
    DB_DATASET = auto()
    UNKNOWN = auto()


class HiveSerdeType(Enum):
    PARQUET = auto()
    AVRO = auto()
    ORC = auto()
    OTHER_HIVESERDE = auto()
    NOT_HIVESERDE = auto()
    INVALID_HIVESERDE_INFO = auto()


class AclMigrationWhat(Enum):
    LEGACY_TACL = auto()
    PRINCIPAL = auto()


@dataclass
class Table:  # pylint: disable=too-many-public-methods
    catalog: str
    database: str
    name: str
    object_type: str
    table_format: str

    location: str | None = None
    view_text: str | None = None
    # really means migrated_to
    upgraded_to: str | None = None

    storage_properties: str | None = None
    is_partitioned: bool = False

    __id_attributes__: typing.ClassVar[tuple[str, ...]] = ("catalog", "database", "name")

    DBFS_ROOT_PREFIXES: typing.ClassVar[list[str]] = [
        "/dbfs/",
        "dbfs:/",
    ]

    DBFS_ROOT_PREFIX_EXCEPTIONS: typing.ClassVar[list[str]] = [
        "/dbfs/mnt",
        "dbfs:/mnt",
    ]

    DBFS_DATABRICKS_DATASETS: typing.ClassVar[list[str]] = [
        "/dbfs/databricks-datasets",
        "dbfs:/databricks-datasets",
    ]

    UPGRADED_FROM_WS_PARAM: typing.ClassVar[str] = "upgraded_from_workspace_id"

    def __post_init__(self) -> None:
        if isinstance(self.table_format, str):  # Should not happen according to type hint, still safer
            self.table_format = self.table_format.upper()

    @property
    def is_delta(self) -> bool:
        if self.table_format is None:
            return False
        return self.table_format.upper() == "DELTA"

    @property
    def is_hive(self) -> bool:
        if self.table_format is None:
            return False
        return self.table_format.upper() == "HIVE"

    @property
    def key(self) -> str:
        # TODO: https://github.com/databrickslabs/ucx/issues/2979
        if self.is_table_in_mount:
            return f"{self.catalog}.{self.database}.{self.location}".lower()
        return f"{self.catalog}.{self.database}.{self.name}".lower()

    @property
    def safe_sql_key(self) -> str:
        return escape_sql_identifier(self.key)

    @property
    def full_name(self) -> str:
        return f"{self.catalog}.{self.database}.{self.name}"

    def __hash__(self):
        return hash(self.key)

    def __eq__(self, other):
        return isinstance(other, Table) and self.key == other.key

    @property
    def kind(self) -> str:
        return "VIEW" if self.view_text is not None else "TABLE"

    def sql_unset_upgraded_to(self):
        return f"ALTER {self.kind} {escape_sql_identifier(self.key)} UNSET TBLPROPERTIES IF EXISTS('upgraded_to');"

    @property
    def is_dbfs_root(self) -> bool:
        if not self.location:
            return False
        for prefix in self.DBFS_ROOT_PREFIXES:
            if not self.location.startswith(prefix):
                continue
            for exception in self.DBFS_ROOT_PREFIX_EXCEPTIONS:
                if self.location.startswith(exception):
                    return False
            for db_datasets in self.DBFS_DATABRICKS_DATASETS:
                if self.location.startswith(db_datasets):
                    return False
            return True
        return False

    @property
    def is_dbfs_mnt(self) -> bool:
        if not self.location:
            return False
        for dbfs_mnt_prefix in self.DBFS_ROOT_PREFIX_EXCEPTIONS:
            if self.location.startswith(dbfs_mnt_prefix):
                return True
        return False

    @property
    def is_format_supported_for_sync(self) -> bool:
        if self.table_format is None:
            return False
        return self.table_format.upper() in {"DELTA", "PARQUET", "CSV", "JSON", "ORC", "TEXT", "AVRO"}

    @property
    def is_databricks_dataset(self) -> bool:
        if not self.location:
            return False
        for db_datasets in self.DBFS_DATABRICKS_DATASETS:
            if self.location.startswith(db_datasets):
                return True
        return False

    @property
    def is_table_in_mount(self) -> bool:
        return self.database.startswith("mounted_") and self.is_delta

    @property
    def is_managed(self) -> bool:
        return self.object_type == "MANAGED"

    @property
    def what(self) -> What:
        if self.is_databricks_dataset:
            return What.DB_DATASET
        if self.is_table_in_mount:
            return What.TABLE_IN_MOUNT
        if self.is_dbfs_root and self.is_delta:
            return What.DBFS_ROOT_DELTA
        if self.is_dbfs_root:
            return What.DBFS_ROOT_NON_DELTA
        if self.kind == "TABLE" and self.is_format_supported_for_sync:
            return What.EXTERNAL_SYNC
        if self.kind == "TABLE" and self.is_hive:
            return What.EXTERNAL_HIVESERDE
        if self.kind == "TABLE":
            return What.EXTERNAL_NO_SYNC
        if self.kind == "VIEW":
            return What.VIEW
        return What.UNKNOWN

    def sql_migrate_external(self, target_table_key):
        return f"SYNC TABLE {escape_sql_identifier(target_table_key)} FROM {escape_sql_identifier(self.key)};"

    def sql_migrate_as_external(self, target_table_key):
        return (
            f"SYNC TABLE {escape_sql_identifier(target_table_key)} AS EXTERNAL FROM {escape_sql_identifier(self.key)};"
        )

    def sql_migrate_ctas_external(self, target_table_key, dst_table_location) -> str:
        return (
            f"CREATE TABLE IF NOT EXISTS {escape_sql_identifier(target_table_key)} "
            f"LOCATION '{dst_table_location}' "
            f"AS SELECT * FROM {self.safe_sql_key}"
        )

    def sql_migrate_ctas_managed(self, target_table_key) -> str:
        return (
            f"CREATE TABLE IF NOT EXISTS {escape_sql_identifier(target_table_key)} "
            f"AS SELECT * FROM {self.safe_sql_key}"
        )

    def hiveserde_type(self, backend: SqlBackend) -> HiveSerdeType:
        if not self.is_hive:
            return HiveSerdeType.NOT_HIVESERDE
        # Extract hive serde info, ideally this should be done by table crawler.
        # But doing here to avoid breaking change to the `tables` table in the inventory schema.
        describe = {}
        for key, values, _ in backend.fetch(f"DESCRIBE TABLE EXTENDED {escape_sql_identifier(self.key)}"):
            describe[key] = values
        if not {"Serde Library", "InputFormat", "OutputFormat"} <= describe.keys():
            return HiveSerdeType.INVALID_HIVESERDE_INFO
        serde = describe["Serde Library"]
        input_format = describe["InputFormat"]
        output_format = describe["OutputFormat"]
        if self._if_parquet_serde(serde, input_format, output_format):
            return HiveSerdeType.PARQUET
        if self._if_avro_serde(serde, input_format, output_format):
            return HiveSerdeType.AVRO
        if self._if_orc_serde(serde, input_format, output_format):
            return HiveSerdeType.ORC
        return HiveSerdeType.OTHER_HIVESERDE

    def _if_parquet_serde(self, serde, input_format, output_format) -> bool:
        return (
            serde == "org.apache.hadoop.hive.ql.io.parquet.serde.ParquetHiveSerDe"
            and input_format == "org.apache.hadoop.hive.ql.io.parquet.MapredParquetInputFormat"
            and output_format == "org.apache.hadoop.hive.ql.io.parquet.MapredParquetOutputFormat"
        )

    def _if_avro_serde(self, serde, input_format, output_format) -> bool:
        return (
            serde == "org.apache.hadoop.hive.serde2.avro.AvroSerDe"
            and input_format == "org.apache.hadoop.hive.ql.io.avro.AvroContainerInputFormat"
            and output_format == "org.apache.hadoop.hive.ql.io.avro.AvroContainerOutputFormat"
        )

    def _if_orc_serde(self, serde, input_format, output_format) -> bool:
        return (
            serde == "org.apache.hadoop.hive.ql.io.orc.OrcSerde"
            and input_format == "org.apache.hadoop.hive.ql.io.orc.OrcInputFormat"
            and output_format == "org.apache.hadoop.hive.ql.io.orc.OrcOutputFormat"
        )

    def sql_migrate_external_hiveserde_in_place(
        self,
        catalog_name,
        dst_schema,
        dst_table,
        backend: SqlBackend,
        hiveserde_type: HiveSerdeType,
        replace_table_location: str | None = None,
    ) -> str | None:
        # "PARQUET", "AVRO", "ORC" can be migrated with "SHOW CREATE TABLE..." DDL
        if hiveserde_type in [HiveSerdeType.PARQUET, HiveSerdeType.AVRO, HiveSerdeType.ORC]:
            return self._ddl_show_create_table(backend, catalog_name, dst_schema, dst_table, replace_table_location)

        # "TEXTFILE" hiveserde needs extra handling on preparing the DDL
        # TODO: add support for "TEXTFILE" hiveserde, when the data can be parsed as Spark CSV datasource

        # "JSON", "CSV" hiveserde need extra handling on preparing the DDL
        # TODO: DBR does not bundle the jars for "JSON", "CSV" hiveserde, it's unlikely we see those tables.
        #  Although it's possible that users has the jars installed as cluster library and use those tables in Databricks,
        #  we hold off the implementation for now until we see the real use case.
        return None

    def _ddl_show_create_table(
        self, backend: SqlBackend, catalog_name, dst_schema, dst_table, replace_location
    ) -> str | None:
        # get raw DDL from "SHOW CREATE TABLE..."
        createtab_stmt = next(backend.fetch(f"SHOW CREATE {self.kind} {self.safe_sql_key};"))["createtab_stmt"]
        # parse the DDL and replace the old table name with the new UC table name
        try:
            statements = sqlglot.parse(createtab_stmt)
        except (ValueError, ParseError):
            logger.exception(f"Exception when parsing 'SHOW CREATE TABLE' DDL for {self.key}")
            return None

        statement = statements[0]
        if not statement:
            logger.error(f"sqlglot parsed none statement from 'SHOW CREATE TABLE' DDL for {self.key}")
            return None

        src_table = statement.find(expressions.Table)
        if not src_table:
            logger.error(f"sqlglot failed to extract table object from parsed DDL for {self.key}")
            return None
        new_table = expressions.Table(catalog=catalog_name, db=dst_schema, this=dst_table)
        src_table.replace(new_table)

        if replace_location:
            # replace dbfs mnt in ddl if any
            mnt_loc = statement.find(LocationProperty)
            if not mnt_loc:
                logger.error(f"sqlglot failed to extract table location object from parsed DDL for {self.key}")
                return None
            new_loc = LocationProperty(this=f"'{replace_location}'")
            mnt_loc.replace(new_loc)

        new_sql = statement.sql('databricks')
        return new_sql

    def sql_migrate_dbfs(self, target_table_key):
        if not self.is_delta:
            msg = f"{self.key} is not DELTA: {self.table_format}"
            raise ValueError(msg)
        return f"CREATE TABLE IF NOT EXISTS {escape_sql_identifier(target_table_key)} DEEP CLONE {escape_sql_identifier(self.key)};"

    def sql_migrate_view(self, target_table_key):
        return f"CREATE VIEW IF NOT EXISTS {escape_sql_identifier(target_table_key)} AS {self.view_text};"

    def sql_migrate_table_in_mount(self, target_table_key: str, table_schema: Iterator[typing.Any]):
        fields = []
        partitioned_fields = []
        next_fields_are_partitioned = False
        for key, value, _ in table_schema:
            if key == "# Partition Information":
                continue
            if key == "# col_name":
                next_fields_are_partitioned = True
                continue
            if next_fields_are_partitioned:
                partitioned_fields.append(escape_sql_identifier(key, maxsplit=0))
            else:
                fields.append(f"{escape_sql_identifier(key, maxsplit=0)} {value}")

        partitioned_str = ""
        if partitioned_fields:
            partitioning_columns = ", ".join(partitioned_fields)
            partitioned_str = f"PARTITIONED BY ({partitioning_columns})"
        schema = ", ".join(fields)

        return f"CREATE TABLE IF NOT EXISTS {escape_sql_identifier(target_table_key)} ({schema}) {partitioned_str} LOCATION '{self.location}';"


@dataclass
class TableError:
    catalog: str
    database: str
    name: str | None = None
    error: str | None = None


@dataclass
class MigrationCount:
    database: str
    what_count: dict[What, int]


class TablesCrawler(CrawlerBase[Table]):
    def __init__(self, sql_backend: SqlBackend, schema, include_databases: list[str] | None = None):
        """
        Initializes a TablesCrawler instance.

        Args:
            sql_backend (SqlBackend): The SQL Execution Backend abstraction (either REST API or Spark)
            schema: The schema name for the inventory persistence.
        """
        super().__init__(sql_backend, "hive_metastore", schema, "tables", Table)
        self._include_database = include_databases

    def _all_databases(self) -> list[str]:
        if not self._include_database:
            databases = []
            for row in self._fetch("SHOW DATABASES"):
                database = row[0]
                if database == self._schema:
                    logger.debug(f"Skipping UCX inventory schema: {database}")
                    continue
                databases.append(database)
            return databases
        return self._include_database

    def load_one(self, schema_name: str, table_name: str) -> Table | None:
        query = f"SELECT * FROM {escape_sql_identifier(self.full_name)} WHERE database='{schema_name}' AND name='{table_name}' LIMIT 1"
        for row in self._fetch(query):
            return Table(*row)
        return None

    @staticmethod
    def _parse_table_props(tbl_props: str) -> dict:
        pattern = r"([^,\[\]]+)=([^,\[\]]+)"
        key_value_pairs = re.findall(pattern, tbl_props)
        # Convert key-value pairs to dictionary
        return dict(key_value_pairs)

    @staticmethod
    def parse_database_props(tbl_props: str) -> dict:
        pattern = r"([^,^\(^\)\[\]]+),([^,^\(^\)\[\]]+)"
        key_value_pairs = re.findall(pattern, tbl_props)
        # Convert key-value pairs to dictionary
        return dict(key_value_pairs)

    def _try_fetch(self) -> Iterable[Table]:
        """Tries to load table information from the database or throws TABLE_OR_VIEW_NOT_FOUND error"""
        for row in self._fetch(f"SELECT * FROM {escape_sql_identifier(self.full_name)}"):
            yield Table(*row)

    def _show_tables_in_database(self, catalog: str, database: str) -> list[Table]:
        table_rows: list[Table] = []
        try:
            logger.debug(f"[{catalog}.{database}] listing tables and views")
            for row in self._fetch(
                f"SHOW TABLES FROM {escape_sql_identifier(catalog)}.{escape_sql_identifier(database)}"
            ):
                table_rows.append(
                    Table(
                        catalog=catalog,
                        database=database,
                        name=row[1],
                        object_type='UNKNOWN',
                        table_format='UNKNOWN',
                    )
                )
        except NotFound:
            # This make the integration test more robust as many test schemas are being created and deleted quickly.
            # In case a schema is deleted, StatementExecutionBackend returns empty result but RuntimeBackend
            # raises NotFound
            logger.warning(f"Schema {catalog}.{database} no longer exists")
        return table_rows

    def _crawl(self) -> Iterable[Table]:
        """Crawls and lists tables within the specified catalog and database.

        After performing initial scan of all tables, starts making parallel
        DESCRIBE TABLE EXTENDED queries for every table.

        Production tasks would most likely be executed through FasterTableScanCrawler
        within `crawl_tables` task due to `spark.sharedState.externalCatalog`
        lower-level APIs not requiring a roundtrip to storage, which is not
        possible for Azure storage with credentials supplied through Spark
        conf (see https://github.com/databrickslabs/ucx/issues/249).

        FasterTableScanCrawler uses the _jsparkSession to utilize faster scanning with Scala APIs.

        See also https://github.com/databrickslabs/ucx/issues/247
        """
        tasks = []
        catalog = "hive_metastore"
        table_names = [partial(self._show_tables_in_database, catalog, database) for database in self._all_databases()]
        for batch in Threads.strict('listing tables', table_names):
            for table in batch:
                tasks.append(partial(self._describe, table.catalog, table.database, table.name))
        catalog_tables, errors = Threads.gather(f"describing tables in {catalog}", tasks)
        if len(errors) > 0:
            logger.error(f"Detected {len(errors)} while scanning tables in {catalog}")
        return catalog_tables

    @staticmethod
    def _safe_norm(value: str | None, *, lower: bool = True) -> str | None:
        if not value:
            return None
        if lower:
            return value.lower()
        return value.upper()

    def _describe(self, catalog: str, database: str, table: str) -> Table | None:
        """Fetches metadata like table type, data format, external table location,
        and the text of a view if specified for a specific table within the given
        catalog and database.
        """
        full_name = f"{catalog}.{database}.{table}"
        try:
            logger.debug(f"[{full_name}] fetching table metadata")
            describe = {}
            for key, value, _ in self._fetch(f"DESCRIBE TABLE EXTENDED {escape_sql_identifier(full_name)}"):
                describe[key] = value
            return Table(
                catalog=catalog.lower(),
                database=database.lower(),
                name=table.lower(),
                object_type=describe.get("Type", "UNKNOWN").upper(),
                table_format=describe.get("Provider", "UNKNOWN").upper(),
                location=describe.get("Location", None),
                view_text=describe.get("View Text", None),
                upgraded_to=self._parse_table_props(describe.get("Table Properties", "").lower()).get(
                    "upgraded_to", None
                ),
                storage_properties=self._parse_table_props(
                    describe.get("Storage Properties", "").lower()
                ),  # type: ignore[arg-type]
                is_partitioned="# Partition Information" in describe,
            )
        except NotFound:
            # This make the integration test more robust as many test schemas are being created and deleted quickly.
            logger.warning(f"Schema {catalog}.{database} no longer existed")
            return None
        except Exception as e:  # pylint: disable=broad-exception-caught
            logger.error(f"Couldn't fetch information for table {full_name} : {e}")
            return None


class FasterTableScanCrawler(TablesCrawler):
    """
    FasterTableScanCrawler is a specialized version of TablesCrawler that uses spark._jsparkSession to utilize
    faster scanning with Scala APIs.

    For integration testing, FasterTableScanCrawler is tested using the larger assessment test rather than
    just the class. Testing the class individually would require utilizing a remote spark connection with the
    Databricks workspace.
    """

    def __init__(self, sql_backend: SqlBackend, schema, include_databases: list[str] | None = None):
        self._sql_backend = sql_backend
        self._include_database = include_databases

        # pylint: disable-next=import-error,import-outside-toplevel
        from pyspark.sql.session import SparkSession  # type: ignore[import-not-found]

        super().__init__(sql_backend, schema, include_databases)
        self._spark = SparkSession.builder.getOrCreate()

    @cached_property
    def _external_catalog(self):
        return self._spark._jsparkSession.sharedState().externalCatalog()  # pylint: disable=protected-access

    def _iterator(self, result: typing.Any) -> Iterator:
        iterator = result.iterator()
        while iterator.hasNext():
            yield iterator.next()

    @staticmethod
    def _option_as_python(scala_option: typing.Any):
        return scala_option.get() if scala_option.isDefined() else None

    def _all_databases(self) -> list[str]:
        if self._include_database:
            return self._include_database
        try:
            return list(self._iterator(self._external_catalog.listDatabases()))
        except Exception as err:  # pylint: disable=broad-exception-caught
            if "py4j.security.Py4JSecurityException" in str(err):
                logger.error(
                    "Failed to list databases due to Py4JSecurityException. "
                    "Update or reinstall UCX to resolve this issue.",
                    exc_info=True,
                )
                return []
            logger.error(f"failed-table-crawl: listing databases -> catalog : {err}", exc_info=True)
            return []

    def _list_tables(self, database: str) -> list[str]:
        try:
            return list(self._iterator(self._external_catalog.listTables(database)))
        except Exception as err:  # pylint: disable=broad-exception-caught
            if "py4j.security.Py4JSecurityException" in str(err):
                logger.error(
                    "Failed to list databases due to Py4JSecurityException. "
                    "Update or reinstall UCX to resolve this issue.",
                    exc_info=True,
                )
                return []
            logger.warning(f"failed-table-crawl: listing tables from database -> {database} : {err}", exc_info=True)
            return []

    def _try_fetch(self) -> Iterable[Table]:
        """Tries to load table information from the database or throws TABLE_OR_VIEW_NOT_FOUND error"""
        for row in self._fetch(f"SELECT * FROM {escape_sql_identifier(self.full_name)}"):
            yield Table(*row)

    @staticmethod
    def _format_properties_list(properties_list: list) -> str:
        if len(properties_list) == 0:
            return ""
        formatted_items: list[str] = []
        for each_property in properties_list:
            key = each_property.productElement(0)
            value = each_property.productElement(1)

            redacted_key = "*******"

            if key == "personalAccessToken" or key.lower() == "password":
                value = redacted_key
            elif value is None:
                value = "None"

            formatted_items.append(f"{key}={value}")
        return f"[{', '.join(formatted_items)}]"

    def _describe(self, catalog, database, table) -> Table | None:
        """Fetches metadata like table type, data format, external table location,
        and the text of a view if specified for a specific table within the given
        catalog and database.
        """
        full_name = f"{catalog}.{database}.{table}"
        if catalog != "hive_metastore":
            msg = f"Only tables in the hive_metastore catalog can be described: {full_name}"
            raise ValueError(msg)
        logger.debug(f"Fetching metadata for table: {full_name}")
        try:  # pylint: disable=too-many-try-statements
            raw_table = self._external_catalog.getTable(database, table)
            table_format = self._option_as_python(raw_table.provider()) or "UNKNOWN"
            location_uri = self._option_as_python(raw_table.storage().locationUri())
            if location_uri:
                location_uri = location_uri.toString()
            is_partitioned = raw_table.partitionColumnNames().iterator().hasNext()
            object_type = raw_table.tableType().name()
            view_text = self._option_as_python(raw_table.viewText())
            table_properties = list(self._iterator(raw_table.properties()))
            formatted_table_properties = self._format_properties_list(table_properties)
        except Exception as e:  # pylint: disable=broad-exception-caught
            logger.warning(f"failed-table-crawl: describing table -> {full_name}: {e}", exc_info=True)
            return None
        return Table(
            catalog=catalog,
            database=database,
            name=table,
            object_type=object_type,
            table_format=table_format,
            location=location_uri,
            view_text=view_text,
            storage_properties=formatted_table_properties,
            is_partitioned=is_partitioned,
        )

    def _crawl(self) -> Iterable[Table]:
        """Crawls and lists tables within the specified catalog and database."""
        tasks = []
        catalog_tables: Collection[Table]
        catalog = "hive_metastore"
        databases = self._all_databases()
        for database in databases:
            if database == self._schema:
                logger.debug(f"Skipping UCX inventory schema: {database}")
                continue
            logger.info(f"Scanning {database}")
            table_names = self._get_table_names(database)
            tasks.extend(self._create_describe_tasks(catalog, database, table_names))
        catalog_tables, errors = Threads.gather("describing tables in ", tasks)
        if len(errors) > 0:
            logger.warning(f"Detected {len(errors)} errors while scanning tables in ")

        logger.info(f"Finished scanning {len(catalog_tables)} tables")
        return catalog_tables

    def _get_table_names(self, database: str) -> list[str]:
        """
        Lists tables names in the specified database.
        :param database:
        :return: list of table names
        """
        # TODO: this method is redundant and can be removed in favor of using _list_tables directly
        table_names = []
        table_names_batches = Threads.strict('listing tables', [partial(self._list_tables, database)])
        for table_batch in table_names_batches:
            table_names.extend(table_batch)
        return table_names

    def _create_describe_tasks(self, catalog: str, database: str, table_names: list[str]) -> list[partial]:
        """
        Creates a list of partial functions for describing tables.
        :param catalog:
        :param database:
        :param table_names:
        :return: list of partial functions
        """
        tasks = []
        for table in table_names:
            tasks.append(partial(self._describe, catalog, database, table))
        return tasks



import datetime as dt

from databricks.labs.ucx.contexts.workflow_task import RuntimeContext
from databricks.labs.ucx.framework.tasks import Workflow, job_task


class MigrationProgress(Workflow):
    """Experimental workflow that rescans the environment to reflect and track progress that has been made.

    It overlaps substantially with the assessment workflow, covering:

     - Clusters
     - Dashboards
     - Grants
     - Jobs (inventory & linting)
     - Pipelines
     - Policies
     - Tables
     - TableMigrationStatus
     - UDFs
    """

    def __init__(self) -> None:
        super().__init__('migration-progress-experimental')

    @job_task(job_cluster="user_isolation")
    def verify_prerequisites(self, ctx: RuntimeContext) -> None:
        """Verify the prerequisites for running this job on the table migration cluster are fulfilled.

        We will wait up to 1 hour for the assessment run to finish if it is running or pending.
        """
        ctx.verify_progress_tracking.verify(timeout=dt.timedelta(hours=1))

    @job_task(depends_on=[verify_prerequisites])
    def crawl_tables(self, ctx: RuntimeContext) -> None:
        """Iterates over all tables in the Hive Metastore of the current workspace and persists their metadata, such
        as _database name_, _table name_, _table type_, _table location_, etc., in the table named
        `$inventory_database.tables`. The metadata stored is then used in the subsequent tasks and workflows to, for
        example, find all Hive Metastore tables that cannot easily be migrated to Unity Catalog."""
        # The table inventory cannot be (quickly) crawled from the table_migration cluster, and the main cluster is not
        # UC-enabled, so we cannot both snapshot and update the history log from the same location.
        # Step 1 of 3: Just refresh the inventory.
        ctx.tables_crawler.snapshot(force_refresh=True)

    @job_task(depends_on=[verify_prerequisites, crawl_tables], job_cluster="user_isolation")
    def refresh_table_migration_status(self, ctx: RuntimeContext) -> None:
        """Scan the tables (and views) in the inventory and record whether each has been migrated or not."""
        # Step 2 of 3: Refresh the migration status of all the tables (updated in the previous step on the main cluster.)
        ctx.migration_status_refresher.snapshot(force_refresh=True)

    @job_task(
        depends_on=[verify_prerequisites, crawl_tables, refresh_table_migration_status], job_cluster="user_isolation"
    )
    def update_tables_history_log(self, ctx: RuntimeContext) -> None:
        """Update the history log with the latest tables inventory snapshot."""
        # Step 3 of 3: Assuming (due to depends-on) the inventory and migration status were refreshed, capture into the
        # history log.
        # TODO: Avoid triggering implicit refresh here if either the table or migration-status inventory is empty.
        history_log = ctx.tables_progress
        tables_snapshot = ctx.tables_crawler.snapshot()
        history_log.append_inventory_snapshot(tables_snapshot)

    @job_task(depends_on=[verify_prerequisites], job_cluster="user_isolation")
    def crawl_udfs(self, ctx: RuntimeContext) -> None:
        """Iterates over all UDFs in the Hive Metastore of the current workspace and persists their metadata in the
        table named `$inventory_database.udfs`. This inventory is currently used when scanning securable objects for
        issues with grants that cannot be migrated to Unit Catalog."""
        history_log = ctx.udfs_progress
        udfs_snapshot = ctx.udfs_crawler.snapshot(force_refresh=True)
        history_log.append_inventory_snapshot(udfs_snapshot)

    @job_task(depends_on=[verify_prerequisites, crawl_tables, crawl_udfs], job_cluster="user_isolation")
    def crawl_grants(self, ctx: RuntimeContext) -> None:
        """Scans all securable objects for permissions that have been assigned: this include database-level permissions,
        as well permissions directly configured on objects in the (already gathered) table and UDF inventories. The
        captured information is stored in the `$inventory_database.grants` inventory table for further use during the
        migration of legacy ACLs to Unity Catalog permissions.

        Note: This job runs on a separate cluster (named `tacl`) as it requires the proper configuration to have the Table
        ACLs enabled and available for retrieval."""
        history_log = ctx.grants_progress
        grants_snapshot = ctx.grants_crawler.snapshot(force_refresh=True)
        history_log.append_inventory_snapshot(grants_snapshot)

    @job_task(depends_on=[verify_prerequisites], job_cluster="user_isolation")
    def assess_jobs(self, ctx: RuntimeContext) -> None:
        """Scans through all the jobs and identifies those that are not compatible with UC. The list of all the jobs is
        stored in the `$inventory.jobs` table.

        It looks for:
          - Clusters with Databricks Runtime (DBR) version earlier than 11.3
          - Clusters using Passthrough Authentication
          - Clusters with incompatible Spark config tags
          - Clusters referencing DBFS locations in one or more config options
        """
        history_log = ctx.jobs_progress
        jobs_snapshot = ctx.jobs_crawler.snapshot(force_refresh=True)
        history_log.append_inventory_snapshot(jobs_snapshot)

    @job_task(depends_on=[verify_prerequisites], job_cluster="user_isolation")
    def assess_clusters(self, ctx: RuntimeContext) -> None:
        """Scan through all the clusters and identifies those that are not compatible with UC. The list of all the clusters
        is stored in the`$inventory.clusters` table.

        It looks for:
          - Clusters with Databricks Runtime (DBR) version earlier than 11.3
          - Clusters using Passthrough Authentication
          - Clusters with incompatible spark config tags
          - Clusters referencing DBFS locations in one or more config options
        """
        history_log = ctx.clusters_progress
        clusters_snapshot = ctx.clusters_crawler.snapshot(force_refresh=True)
        history_log.append_inventory_snapshot(clusters_snapshot)

    @job_task(depends_on=[verify_prerequisites], job_cluster="user_isolation")
    def assess_pipelines(self, ctx: RuntimeContext) -> None:
        """This module scans through all the Pipelines and identifies those pipelines which has Azure Service Principals
        embedded (who has been given access to the Azure storage accounts via spark configurations) in the pipeline
        configurations.

        It looks for:
          - all the pipelines which has Azure Service Principal embedded in the pipeline configuration

        Subsequently, a list of all the pipelines with matching configurations are stored in the
        `$inventory.pipelines` table."""
        history_log = ctx.pipelines_progress
        pipelines_snapshot = ctx.pipelines_crawler.snapshot(force_refresh=True)
        history_log.append_inventory_snapshot(pipelines_snapshot)

    @job_task(depends_on=[verify_prerequisites], job_cluster="user_isolation")
    def crawl_cluster_policies(self, ctx: RuntimeContext) -> None:
        """This module scans through all the Cluster Policies and get the necessary information

        It looks for:
          - Clusters Policies with Databricks Runtime (DBR) version earlier than 11.3

          Subsequently, a list of all the policies with matching configurations are stored in the
        `$inventory.policies` table."""
        history_log = ctx.policies_progress
        cluster_policies_snapshot = ctx.policies_crawler.snapshot(force_refresh=True)
        history_log.append_inventory_snapshot(cluster_policies_snapshot)

    @job_task(depends_on=[verify_prerequisites])
    def assess_dashboards(self, ctx: RuntimeContext):
        """Scans all dashboards for migration issues in SQL code of embedded widgets.
        Also stores direct filesystem accesses for display in the migration dashboard."""
        # TODO: Ensure these are captured in the history log.
        ctx.query_linter.refresh_report()

    @job_task(depends_on=[verify_prerequisites])
    def assess_workflows(self, ctx: RuntimeContext):
        """Scans all jobs for migration issues in notebooks.
        Also stores direct filesystem accesses for display in the migration dashboard."""
        # TODO: Ensure these are captured in the history log.
        ctx.workflow_linter.refresh_report(ctx.sql_backend, ctx.inventory_database)

    @job_task(
        depends_on=[
            verify_prerequisites,
            crawl_grants,
            crawl_udfs,
            assess_jobs,
            assess_clusters,
            assess_pipelines,
            crawl_cluster_policies,
            refresh_table_migration_status,
            update_tables_history_log,
        ],
        job_cluster="user_isolation",
    )
    def record_workflow_run(self, ctx: RuntimeContext) -> None:
        """Record the workflow run of this workflow."""
        ctx.workflow_run_recorder.record()



from __future__ import annotations

import heapq
import itertools
from collections import defaultdict
from collections.abc import Iterable
from dataclasses import dataclass, field

from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import DatabricksError
from databricks.sdk.service.jobs import Job, JobCluster, Task

from databricks.labs.ucx.assessment.clusters import ClusterOwnership, ClusterInfo
from databricks.labs.ucx.assessment.jobs import JobOwnership, JobInfo
from databricks.labs.ucx.framework.owners import AdministratorLocator
from databricks.labs.ucx.source_code.graph import DependencyProblem


@dataclass
class MigrationStep:
    step_id: int
    """Globally unique id."""

    step_number: int
    """The position in the migration sequence."""

    object_type: str
    """Object type. Together with `attr:object_id` a unique identifier."""

    object_id: str
    """Object id. Together with `attr:object_id` a unique identifier."""

    object_name: str
    """Object name, more human friendly than `attr:object_id`."""

    object_owner: str
    """Object owner."""

    required_step_ids: list[int]
    """The step ids that should be completed before this step is started."""


MigrationNodeKey = tuple[str, str]


@dataclass(frozen=True)
class MigrationNode:
    node_id: int = field(compare=False)
    """Globally unique id."""

    object_type: str
    """Object type. Together with `attr:object_id` a unique identifier."""

    object_id: str
    """Object id. Together with `attr:object_id` a unique identifier."""

    object_name: str = field(compare=False)
    """Object name, more human friendly than `attr:object_id`."""

    object_owner: str = field(compare=False)
    """Object owner."""

    @property
    def key(self) -> MigrationNodeKey:
        """Unique identifier of the node."""
        return self.object_type, self.object_id

    def as_step(self, step_number: int, required_step_ids: list[int]) -> MigrationStep:
        """Convert to class:MigrationStep."""
        return MigrationStep(
            step_id=self.node_id,
            step_number=step_number,
            object_type=self.object_type,
            object_id=self.object_id,
            object_name=self.object_name,
            object_owner=self.object_owner,
            required_step_ids=required_step_ids,
        )


@dataclass
class MaybeMigrationNode:
    node: MigrationNode | None
    problems: list[DependencyProblem]

    @property
    def failed(self) -> bool:
        return len(self.problems) > 0


# We expect `tuple[int, int, MigrationNode | str]`
# for `[priority, counter, MigrationNode | PriorityQueue._REMOVED | PriorityQueue_UPDATED]`
# but we use list for the required mutability
QueueEntry = list[int | MigrationNode | str]


class PriorityQueue:
    """A migration node priority queue.

    Note:
        This implementation does not support threading safety as that is not required.

    Source:
        See https://docs.python.org/3/library/heapq.html#priority-queue-implementation-notes on the changes below
        to handle priority changes in the task. Also, the _UPDATED marker is introduced to avoid updating removed nodes.
    """

    _REMOVED = "<removed>"  # Mark removed items

    def __init__(self):
        self._entries: list[QueueEntry] = []
        self._entry_finder: dict[MigrationNode, QueueEntry] = {}
        self._counter = itertools.count()  # Tiebreaker with equal priorities, then "first in, first out"

    def put(self, priority: int, task: MigrationNode) -> None:
        """Put or update task in the queue.

        The lowest priority is retrieved from the queue first.
        """
        if task in self._entry_finder:
            self._remove(task)
        count = next(self._counter)
        entry = [priority, count, task]
        self._entry_finder[task] = entry
        heapq.heappush(self._entries, entry)

    def get(self) -> MigrationNode | None:
        """Gets the tasks with the lowest priority."""
        while self._entries:
            _, _, task = heapq.heappop(self._entries)
            if task == self._REMOVED:
                continue
            assert isinstance(task, MigrationNode)
            self._remove(task)
            return task
        return None

    def _remove(self, task: MigrationNode) -> None:
        """Remove a task from the queue."""
        entry = self._entry_finder.pop(task)
        entry[2] = self._REMOVED


class MigrationSequencer:
    """Sequence the migration dependencies in order to execute the migration.

    Similar to the other graph logic, we first build the graph by registering dependencies, then we analyse the graph.
    Analysing the graph in this case means: computing the migration sequence in `meth:generate_steps`.
    """

    def __init__(self, ws: WorkspaceClient, administrator_locator: AdministratorLocator):
        self._ws = ws
        self._admin_locator = administrator_locator
        self._counter = itertools.count()
        self._nodes: dict[MigrationNodeKey, MigrationNode] = {}

        # Outgoing references contains edges in the graph pointing from a node to a set of nodes that the node
        # references. These references follow the API references, e.g. a job contains tasks in the
        # `Job.settings.tasks`, thus a job has an outgoing reference to each of those tasks.
        self._outgoing_references: dict[MigrationNodeKey, set[MigrationNode]] = defaultdict(set)

    def register_jobs(self, *jobs: Job) -> list[MaybeMigrationNode]:
        """Register a job.

        Args:
            jobs (Job) : The jobs to register.

        Returns:
            list[MaybeMigrationNode] : Each element contains a maybe migration node for each job respectively. If no
                problems occurred during registering the job, the maybe migration node contains the migration node.
                Otherwise, the maybe migration node contains the dependency problems occurring during registering the
                job.
        """
        nodes: list[MaybeMigrationNode] = []
        for job in jobs:
            node = self._register_job(job)
            nodes.append(node)
        return nodes

    def _register_job(self, job: Job) -> MaybeMigrationNode:
        """Register a single job."""
        problems: list[DependencyProblem] = []
        job_node = self._nodes.get(("JOB", str(job.job_id)), None)
        if job_node:
            return MaybeMigrationNode(job_node, problems)
        job_name = job.settings.name if job.settings and job.settings.name else str(job.job_id)
        job_node = MigrationNode(
            node_id=next(self._counter),
            object_type="JOB",
            object_id=str(job.job_id),
            object_name=job_name,
            object_owner=JobOwnership(self._admin_locator).owner_of(JobInfo.from_job(job)),
        )
        self._nodes[job_node.key] = job_node
        if not job.settings:
            return MaybeMigrationNode(job_node, problems)
        for job_cluster in job.settings.job_clusters or []:
            maybe_cluster_node = self._register_job_cluster(job_cluster, job_node)
            if maybe_cluster_node.node:
                self._outgoing_references[job_node.key].add(maybe_cluster_node.node)
        for task in job.settings.tasks or []:
            maybe_task_node = self._register_workflow_task(task, job_node)
            problems.extend(maybe_task_node.problems)
            if maybe_task_node.node:
                self._outgoing_references[job_node.key].add(maybe_task_node.node)
        # Only after registering all tasks, we can resolve the task dependencies
        for task in job.settings.tasks or []:
            task_key = ("TASK", f"{job.job_id}/{task.task_key}")
            for task_dependency in task.depends_on or []:
                task_dependency_key = ("TASK", f"{job.job_id}/{task_dependency.task_key}")
                maybe_task_dependency = self._nodes.get(task_dependency_key)
                if maybe_task_dependency:
                    self._outgoing_references[task_key].add(maybe_task_dependency)
                else:
                    # Verified that a job with a task having a depends on referring a non-existing task cannot be
                    # created. However, this code is just in case.
                    problem = DependencyProblem(
                        'task-dependency-not-found', f"Could not find task: {task_dependency_key[1]}"
                    )
                    problems.append(problem)
        return MaybeMigrationNode(job_node, problems)

    def _register_workflow_task(self, task: Task, parent: MigrationNode) -> MaybeMigrationNode:
        """Register a workflow task.

        TODO:
            Handle following Task attributes:
            - for_each_task
            - libraries
            - notebook_task
            - pipeline_task
            - python_wheel_task
            - run_job_task
            - spark_jar_task
            - spark_python_task
            - spark_submit_task
            - sql_task
        """
        problems: list[DependencyProblem] = []
        task_id = f"{parent.object_id}/{task.task_key}"
        task_node = self._nodes.get(("TASK", task_id), None)
        if task_node:
            return MaybeMigrationNode(task_node, problems)
        task_node = MigrationNode(
            node_id=next(self._counter),
            object_type="TASK",
            object_id=task_id,
            object_name=task.task_key,
            object_owner=parent.object_owner,  # No task owner so use parent job owner
        )
        self._nodes[task_node.key] = task_node
        # `task.new_cluster` is not handled because it is part of the task and not a separate node
        if task.existing_cluster_id:
            maybe_cluster_node = self._register_cluster(task.existing_cluster_id)
            problems.extend(maybe_cluster_node.problems)
            if maybe_cluster_node.node:
                self._outgoing_references[task_node.key].add(maybe_cluster_node.node)
        if task.job_cluster_key:
            job_cluster_node = self._nodes.get(("CLUSTER", f"{parent.object_id}/{task.job_cluster_key}"))
            if job_cluster_node:
                self._outgoing_references[task_node.key].add(job_cluster_node)
            else:
                problem = DependencyProblem('cluster-not-found', f"Could not find cluster: {task.job_cluster_key}")
                problems.append(problem)
        return MaybeMigrationNode(task_node, problems)

    def _register_job_cluster(self, cluster: JobCluster, parent: MigrationNode) -> MaybeMigrationNode:
        """Register a job cluster.

        A job cluster is defined within a job and therefore is found when defined on the job by definition.
        """
        # Different jobs can have job clusters with the same key, therefore job id is prepended to ensure uniqueness
        cluster_id = f"{parent.object_id}/{cluster.job_cluster_key}"
        cluster_node = MigrationNode(
            node_id=next(self._counter),
            object_type="CLUSTER",
            object_id=cluster_id,
            object_name=cluster.job_cluster_key,
            object_owner=parent.object_owner,
        )
        self._nodes[cluster_node.key] = cluster_node
        return MaybeMigrationNode(cluster_node, [])

    def _register_cluster(self, cluster_id: str) -> MaybeMigrationNode:
        """Register a cluster.

        TODO
            Handle following Task attributes:
            - init_scripts
            - instance_pool_id (maybe_not)
            - policy_id
            - spec.init_scripts
            - spec.instance_pool_id (maybe not)
            - spec.policy_id
        """
        node_seen = self._nodes.get(("CLUSTER", cluster_id), None)
        if node_seen:
            return MaybeMigrationNode(node_seen, [])
        try:
            details = self._ws.clusters.get(cluster_id)
        except DatabricksError:
            message = f"Could not find cluster: {cluster_id}"
            return MaybeMigrationNode(None, [DependencyProblem('cluster-not-found', message)])
        object_name = details.cluster_name if details and details.cluster_name else cluster_id
        cluster_node = MigrationNode(
            node_id=next(self._counter),
            object_type="CLUSTER",
            object_id=cluster_id,
            object_name=object_name,
            object_owner=ClusterOwnership(self._admin_locator).owner_of(ClusterInfo.from_cluster_details(details)),
        )
        self._nodes[cluster_node.key] = cluster_node
        # TODO register warehouses and policies
        return MaybeMigrationNode(cluster_node, [])

    def generate_steps(self) -> Iterable[MigrationStep]:
        """Generate the migration steps.

        An adapted version of the Kahn topological sort is implemented. The differences are as follows:
        - We want the same step number for all nodes with same dependency depth. Therefore, instead of pushing to a
          queue, we rebuild it once all leaf nodes are processed (these are transient leaf nodes i.e. they only become
          leaf during processing)
        - We handle cyclic dependencies (implemented in PR #3009)
        """
        ordered_steps: list[MigrationStep] = []
        # For updating the priority of steps that depend on other steps
        incoming_references = self._invert_outgoing_to_incoming_references()
        seen = set[MigrationNode]()
        queue = self._create_node_queue(self._outgoing_references)
        node = queue.get()
        while node is not None:
            step = node.as_step(len(ordered_steps), sorted(n.node_id for n in self._outgoing_references[node.key]))
            ordered_steps.append(step)
            seen.add(node)
            # Update the queue priority as if the migration step was completed
            for dependency in incoming_references[node.key]:
                if dependency in seen:
                    continue
                priority = len(self._outgoing_references[dependency.key] - seen)
                queue.put(priority, dependency)
            node = queue.get()
        return ordered_steps

    def _invert_outgoing_to_incoming_references(self) -> dict[MigrationNodeKey, set[MigrationNode]]:
        """Invert the outgoing references to incoming references."""
        result: dict[MigrationNodeKey, set[MigrationNode]] = defaultdict(set)
        for node_key, outgoing_nodes in self._outgoing_references.items():
            for target in outgoing_nodes:
                result[target.key].add(self._nodes[node_key])
        return result

    def _create_node_queue(self, incoming: dict[MigrationNodeKey, set[MigrationNode]]) -> PriorityQueue:
        """Create a priority queue for their nodes using the incoming count as priority.

        A lower number means it is pulled from the queue first, i.e. the key with the lowest number of keys is retrieved
        first.
        """
        priority_queue = PriorityQueue()
        for node_key, node in self._nodes.items():
            priority_queue.put(len(incoming[node_key]), node)
        return priority_queue
